"""
Base class for row-stacked spectra

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""

# TODO: Pilfer the pypeit.DataContainer for this.

# TODO: Add meta, as in DataCube.

import warnings
from IPython import embed

import numpy
from scipy import sparse, interpolate

from astropy.io import fits

try:
    from shapely.ops import cascaded_union
    from shapely.geometry import Point
except ImportError:
    warnings.warn('Could not import shapely!', ImportWarning)

from ..util.bitmask import BitMask
from ..util.pixelmask import SpectralPixelMask
from ..util.sampling import angstroms_per_pixel
from ..util.covariance import Covariance
from ..util.constants import DAPConstants

class RowStackedSpectra:
    r"""
    Base container class for a set of 1D extracted spectra.

    Row-stacked spectra have two axes: The first is just used to
    organize the spectra, whereas the second is the spectral axis.
    That is, one can loop through the spectra by, e.g.::

        for s in spectra:
            plt.plot(s)
            ...
    
    The spectral axis is expected to be the same for all spectra.

    Currently no spectral or "spatial" (e.g. cross-talk) covariance
    is accounted for.

    .. warning::

        The class can be used to construct datacubes from row-stacked
        spectra. In these calculations, the aperture area is used to
        rescale the flux to ensure that it is conserved per unit
        area. While the class allows the aperture area to be
        fiber-specific, it does not yet allow the smoothing kernel
        used in the rectification to also be fiber-specific. Beware
        if you're trying to construct a cube using multiple fiber
        formats!
    
    Derived classes should, in particular, provide the read methods.
    The critical components of the derived classes are:

    .. todo::

        - Fill in these "critical components"
        - How do I abstract how to provide the on-sky aperture?

    Args:
        wave (`numpy.ndarray`_):
            The wavelength vector of *all* spectra. Shape is
            :math:`(N_\lambda,)`.
        flux (`numpy.ndarray`_):
            Data array with the flux for each spectrum. Shape should
            be :math:`(N_{\rm spec}, N_\lambda)`.
        ivar (`numpy.ndarray`_, optional):
            The inverse variance of the flux array. Shape must be the
            same as ``flux``.  If None, errors are ignored.
        mask (`numpy.ndarray`_, optional):
            Mask array for the flux measurements. Can be a boolean
            mask (False=unmasked,good; True=masked,bad) or an integer
            array associated with the provided
            :class:`mangadap.util.bitmask.BitMask` object (see
            ``bitmask``). Shape must be the same as ``flux``.
        bitmask (:class:`mangadap.util.bitmask.BitMask`, optional):
            Object used to select and toggle masked pixels from
            :attr:`mask`. If None, any provided ``mask`` must be a
            boolean array or can be successfully converted to one.
        sres (`numpy.ndarray`_, optional):
            The spectral resolution, :math:`R = \lambda /
            \Delta\lambda`. Can be a single vector with shape
            :math:`(N_\lambda,)`, if the spectral resolution is
            independent of spatial position, or a 2D set of spectral
            resolution vectors with a shape that matches ``flux``. If
            None, spectral resolution is ignored.
        xpos (`numpy.ndarray`_, optional):
            The wavelength-dependent coordinate along the
            right-ascension direction of the on-sky aperture for all
            spectra. Expected to be arcseconds from some fiducial
            coordinate. Coordinates should increase from W to E.
            Shape must be the same as ``flux``.
        ypos (`numpy.ndarray`_, optional):
            The wavelength-dependent coordinate along the declination
            direction of the on-sky aperture for all spectra.
            Expected to be arcseconds from some fiducial coordinate.
            Coordinates should increase from S to N. Shape must be
            the same as ``flux``.
        area (:obj:`float`, `numpy.ndarray`_, optional):
            The fiducial area subtended by each spectral aperture,
            assumed to be wavelength independent. The aperture is
            assumed to be circular. Can provide a single value for
            all fibers or an array with shape :math:`(N_{\rm spec},)`
            that provides fiber-specific values. If None, the area is
            set to unity, which has follow-on effects on the results
            provided by some of the class methods, as described where
            relevant.
        log (:obj:`bool`, optional):
            Flag that the datacube spectral pixels are binned
            logarithmically in wavelength.
        prihdr (`astropy.io.fits.Header`_, optional):
            Primary header read from source fits file. If None,
            instantiated as an empty `astropy.io.fits.Header`_.
        fluxhdr (`astropy.io.fits.Header`_, optional):
            Header specifically for the flux extension of the source
            fits file. If None, set to be a copy of the primary
            header.

    Attributes:
        shape (:obj:`tuple`):
            Shape of the main spectral arrays.
        nwave (:obj:`int`):
            Number of wavelength channels.
        bitmask (:class:`mangadap.util.bitmask.BitMask`):
            Object used to select and toggle masked pixels from
            :attr:`mask`.  Can be None.
        log (:obj:`bool`):
            Flag that the spectral pixels are binned logarithmically
            in wavelength.
        wave (`numpy.ndarray`_):
            Wavelength vector applicable to all spectra.
        flux (`numpy.ndarray`_):
            Flux array.
        ivar (`numpy.ndarray`_):
            Flux inverse variance.  Can be None.
        mask (`numpy.ndarray`_):
            Spectral mask.  Can be None.
        sres (`numpy.ndarray`_):
            Spectral-resolution array.  Can be None.
        xpos (`numpy.ndarray`_):
            The wavelength-dependent coordinate along the
            right-ascension direction of the on-sky aperture for all
            spectra. Expected to be arcseconds from some fiducial
            coordinate.
        ypos (`numpy.ndarray`_):
            The wavelength-dependent coordinate along the declination
            direction of the on-sky aperture for all spectra.
            Expected to be arcseconds from some fiducial coordinate.
        area (`numpy.ndarray`_):
            The fiducial area subtended by each spectral aperture.
            The aperture is assumed to be circular.
        prihdr (`astropy.io.fits.Header`_):
            Primary header for the row-stacked spectra. If not
            provided on instantiation, set to an empty
            `astropy.io.fits.Header`_.
        fluxhdr (`astropy.io.fits.Header`_):
            Header specifically for the flux array. If not provided
            on instantiation, set to be a copy of :attr:`prihdr`.
    """
    def __init__(self, wave, flux, ivar=None, mask=None, bitmask=None, sres=None, xpos=None,
                 ypos=None, area=None, log=True, prihdr=None, fluxhdr=None):

        # Re-order so that axes are x, y, lambda
        self.wave = wave
        self.flux = flux
        self.shape = self.flux.shape
        self.nwave = self.shape[-1]
        if self.wave.size != self.nwave:
            raise ValueError('Wavelength vector does not match flux array.')
        self.log = log
        # TODO: Check the above against the wavelength vector and/or
        # set the value automatically based on the input.
        
        self.ivar = None 
        if ivar is not None:
            self.ivar = ivar
            if self.ivar.shape != self.shape:
                raise ValueError('Inverse variance array has incorrect shape.')

        self.mask = None
        if mask is not None:
            self.mask = mask
            if self.mask.shape != self.shape:
                raise ValueError('Mask array has incorrect shape.')
        self.bitmask = None
        if bitmask is not None:
            if self.mask is None:
                warnings.warn('Bitmask provided but no mask data available.')
            if isinstance(bitmask, BitMask):
                self.bitmask = bitmask
            else:
                warnings.warn('Bitmasks must be BitMask objects; provided bitmask is ignored.')
        if self.bitmask is None:
            # Ensure mask is a boolean array
            self.mask = self.mask.astype(bool)

        self.sres = None
        if sres is not None:
            self.sres = numpy.tile(sres, (self.shape[0],1)) if sres.ndim == 1 else sres
            if self.sres.shape != self.shape:
                raise ValueError('Spectral resolution array has incorrect shape.')

        self.xpos = None 
        if xpos is not None:
            self.xpos = xpos
            if self.xpos.shape != self.shape:
                raise ValueError('X-coordinate array has incorrect shape.')

        self.ypos = None 
        if ypos is not None:
            self.ypos = ypos
            if self.ypos.shape != self.shape:
                raise ValueError('Y-coordinate array has incorrect shape.')

        self.area = numpy.ones(self.nspec, dtype=float) if area is None else numpy.atleast_1d(area)
        if self.area.size == 1:
            self.area = numpy.full(self.nspec, area, dtype=float)
        if self.area.size != self.nspec:
            raise ValueError('Could not construct area for each aperture; provide a single value '
                             'or one value per spectrum.')

        # Allocate attributes for primary and flux array fits headers
        self.prihdr = fits.Header() if prihdr is None else prihdr
        self.fluxhdr = self.prihdr.deepcopy() if fluxhdr is None else fluxhdr

        # Datacube rectification parameters
        self.pixelscale = None
        self.recenter = None
        self.width_buffer = None
        self.xs = None
        self.nx = None
        self.ys = None
        self.ny = None
        self.rect_rlim = None
        self.rect_sigma = None
        self.rect_channel = None
        self.rect_T = None

    @property
    def nspec(self):
        """Number of spectra in the datacube."""
        return self.shape[0]

    # TODO: Consolidate these functions between RowStackedSpectra and DataCube?
    def copy_to_array(self, attr='flux', waverange=None, nbins=None, select_bins=None,
                      missing_bins=None, unique_bins=None):
        r"""
        Return a copy of the selected data array with a limited
        spectral range and/or a selected set of spectra.

        The array size is always :math:`N_{\rm spec} \times N_{\rm
        wavelength}`. See :func:`copy_to_masked_array` for argument
        descriptions.

        .. warning::
            Any masking is ignored in this function call. To
            incorporate the mask use :func:`copy_to_masked_array`.

        Returns:
            `numpy.ndarray`_: A 2D array with a copy of the data from the
            selected attribute.
        """
        masked_data = self.copy_to_masked_array(attr=attr, use_mask=False, waverange=waverange,
                                                nbins=nbins, select_bins=select_bins,
                                                missing_bins=missing_bins, unique_bins=unique_bins)
        # For this approach, the wavelengths masked should be
        # *identical* for all spectra
        nwave = numpy.sum(numpy.invert(numpy.ma.getmaskarray(masked_data)), axis=1)
        # No masking should be present except for the wavelength range
        # meaning that the number of unmasked pixels at all spatial
        # positions should be the same.
        if numpy.any(nwave != nwave[0]):
            raise ValueError('Masking in copy_to_mask should only for the wavelength range.')
        if numpy.all(nwave == 0):
            raise ValueError('Full wavelength range has been masked!')
        # Compression returns a flattened array, so it needs to be
        # reshaped for the new number of (unmasked) wavelength channels
        return masked_data.compressed().reshape(-1,nwave[0])

    def copy_to_masked_array(self, attr='flux', use_mask=True, flag=None, waverange=None,
                             nbins=None, select_bins=None, missing_bins=None, unique_bins=None):
        r"""
        Return a copy of the selected data array with selected
        pixels, wavelength ranges, and/or full spectra masked.

        This is functionally identical to :func:`copy_to_array`,
        except the output format is a `numpy.ma.MaskedArray`_. The
        pixels that are considered to be masked can be specified
        using the `flag` option. The array size is always
        :math:`N_{\rm spec} \times N_{\rm wavelength}`.
        
        Args:
            attr (:obj:`str`, optional):
                The attribute for the returned array. Can be 'flux',
                'ivar', 'sres', 'mask', 'xpos', or 'ypos'. For
                'mask', you're likely better off using
                :func:`copy_to_array`. Strings are always set to be
                lower-case, so capitalization shouldn't matter.
            use_mask (:obj:`bool`, optional):
                Use the internal mask to mask the data. This is
                largely here to allow for :func:`copy_to_array` to
                wrap this function while not applying the internal
                mask.
            waverange (array-like, optional):
                Two-element array with the first and last wavelength
                to include in the computation. Default is to use the
                full wavelength range.
            flag (:obj:`str`, :obj:`list`, optional):
                One or more bitmask flags used to select bad pixels.
                The names *must* be a valid bit name as defined by
                :attr:`bitmask` (see
                :class:`mangadap.util.bitmask.BitMask`). If not
                provided, *any* non-zero mask bit is omitted.
            nbins (:obj:`int`, optional):
                The total number of defined bins. Default is to find
                the maximum number in the unique_bins list.
            select_bins (array-like, optional):
                A boolean array selecting the spectra to return.
            missing_bins (:obj:`list`, optional):
                A list of bin numbers that are missing from the
                output selection and that will be replaced with
                masked spectra. If specified, must also provide
                ``unique_bins``.
            unique_bins (array-like, optional):
                The indices of the bins that have valid spectra. The
                length of this array should match the select_bins
                array.

        Returns:
            `numpy.ma.MaskedArray`_: A 2D array with a copy of the
            data from the selected extension, masked where
            :attr:`mask` is either True or with the selected bitmask
            flags.

        Raises:
            AttributeError:
                Raised if ``attr`` is not a valid attribute.
            ValueError:
                Raised if the unique bins are not specified, but the
                missing bins are, or if the number of unique bins
                does not match the number of spectra in the datacube.
        """
        # Make sure masks and flags make sense
        if flag is not None and self.bitmask is None:
            warnings.warn('No bitmask defined.  Input flags ignored.')

        nspec = self.nspec

        # Create the wavelength mask (will be all true if
        # waverange=None)
        mask = SpectralPixelMask(waverange=waverange).boolean(self.wave, nspec=nspec)

        # Add in any masked data
        if use_mask:
            mask |= self.mask if self.bitmask is None \
                                else self.bitmask.flagged(self.mask, flag=flag)

        # Create the output MaskedArray
        a = numpy.ma.MaskedArray(getattr(self, attr.lower()), mask=mask)

        # Apply any bin selection
        if select_bins is not None:
            a = a[select_bins,:]

        # No missing bins are specified, so return the array
        if missing_bins is None or len(missing_bins) == 0:
            return a

        # Missing bins have been specified, so the unique_bins also need
        # to be specified
        if unique_bins is None:
            raise ValueError('Must specify the unique bins if missing bins are specified.')
        if len(unique_bins) != a.shape[0]:
            raise ValueError('Number of unique bins does not match the number of spectra.')

        # If the number of bins hasn't been provided, construct it based
        # on the unique bins specified
        _nbins = max(unique_bins)+1 if nbins is None else nbins

        # Create the array with the right shape and mask everything
        _a = numpy.ma.MaskedArray(numpy.zeros((_nbins, a.shape[-1]), dtype=a.dtype),
                                  mask=numpy.ones((_nbins, a.shape[-1]), dtype=bool))
        # Copy the data to the correct bins, which also unmasks these
        # pixels
        _a[unique_bins,:] = a
        # Return the masked array; "missing bins" are fully masked
        return _a

    def interpolate_to_match(self, func, fill_value=0.0):
        r"""
        Interpolate a function to match the datacube wavelength sampling.

        Args:
            func (`numpy.ndarray`_):
                Function to linear interpolate. Shape is
                :math:`(N_{\rm wave}, 2)`, where the first column has
                the wavelength and the second has the function to
                interpolate. If None, simply returns a unity vector
                of the correct length.
            fill_value (:obj:`float`, optional):
                The value to use for spectral regions not sampled by
                the provided function.

        Returns:
            `numpy.ndarray`_: The function interpolated (**not**
            resampled) to match :attr:`wave`. Any spectral regions
            not within the wavelength range of the provided function
            is set to 0.
        """
        if func is None:
            return numpy.ones(self.nwave, dtype=float)
        return interpolate.interp1d(func[:,0], func[:,1], bounds_error=False,
                                    fill_value=fill_value, assume_sorted=True)(self.wave)

    def mean_sky_coordinates(self, waverange=None, response_func=None, per_pixel=True, offset=None,
                             flag=None, fluxwgt=False):
        r"""
        Compute the weighted or unweighted mean sky coordinates for
        each spectrum.

        Method cannot be performed if coordinate arrays are not
        available.

        .. warning::

            Flux-weighting the coordinates can produce spurious
            results in low-flux regimes.

        Args:
            waverange (array-like, optional):
                Two-element array with the first and last wavelength
                to include in the computation. Default is to use the
                full wavelength range.
            response_func (array-like, optional):
                A two-column array with the wavelength and
                transmission of a broad-band response function to use
                for the calculation. If None, no wavelength-dependent
                weighting function is used.
            per_pixel (:obj:`bool`, optional):
                When providing a response function, continue to
                calculate the statistics per pixel, as opposed to per
                angstrom.
            offset (:obj:`tuple`, optional):
                A two-tuple with an x,y offset to apply.
            flag (:obj:`str`, :obj:`list`, optional):
                One or more flag names that are considered when
                deciding if a pixel should be masked. The names
                *must* be a valid bit name as defined by
                :attr:`bitmask`.
            fluxwgt (:obj:`bool`, optional):
                Flag to weight by the flux when determining the mean
                coordinates.

        Returns:
            :obj:`tuple`: Two `numpy.ndarray`_ objects with the
            fiducial x and y on-sky positions of each spectrum in
            arcseconds relative to a given center. The shape of the
            two arrays is :math:`(N_{\rm spec},)`.
        """
        _offset = (0,0) if offset is None else offset
        if len(_offset) != 2:
            raise ValueError('Offset should be two numbers for the offset in x and y.')

        xpos = self.copy_to_masked_array(attr='xpos', waverange=waverange, flag=flag) + _offset[0]
        ypos = self.copy_to_masked_array(attr='ypos', waverange=waverange, flag=flag) + _offset[1]
        if fluxwgt:
            flux = self.copy_to_masked_array(waverange=waverange, flag=flag)

        # Set the response function
        dw = numpy.ones(self.nwave, dtype=float) if per_pixel \
                else angstroms_per_pixel(self.wave, log=self.log)
        _response_func = self.interpolate_to_match(response_func)

        # Get the normalization and return the flux- or un-weighted coordinates
        if fluxwgt:
            norm = numpy.ma.sum(flux*_response_func[None,:]*dw[None,:], axis=1)
            return numpy.ma.sum(flux*xpos*_response_func[None,:]*dw[None,:],axis=1)/norm, \
                    numpy.ma.sum(flux*ypos*_response_func[None,:]*dw[None,:],axis=1)/norm

        norm = numpy.sum(numpy.invert(numpy.ma.getmaskarray(xpos))*_response_func[None,:]
                                *dw[None,:], axis=1)
        return numpy.ma.sum(xpos*_response_func[None,:]*dw[None,:],axis=1)/norm, \
                    numpy.ma.sum(ypos*_response_func[None,:]*dw[None,:],axis=1)/norm

    def binned_on_sky_area(self, bin_indx, x=None, y=None, **kwargs):
        r"""
        Compute the on-sky area of a set of binned spectra.  

        The method attempts to calculate the overlapping area
        assuming the spectra were obtained with a set of circular
        fibers with the provided area (see :attr:`area`). If not
        provided in the method call (see ``x``, ``y``), the fiber
        fiducial coordinates used for the calculation are calculated
        using :func:`mean_sky_coordinates`.

        The overlapping area is calculated using the `shapely`_
        python package. If that package is not available, the
        returned area of each bin is simply the sum of the area of
        each fiber (does not account for any overlaps).

        Args:
            bin_indx (array-like):
                An array with size :math:`N_{\rm spec}` that gives
                which spaxels were included in each bin. Valid bins
                have indices of :math:`\geq 0`.
            x (array-like, optional):
                On-sky :math:`x` coordinate. Default is to calculate
                :math:`x` and :math:`y` using
                :func:`mean_sky_coordinates`.
            y (array-like, optional):
                On-sky :math:`y` coordinate. Default is to calculate
                :math:`x` and :math:`y` using
                :func:`mean_sky_coordinates`.
            **kwargs:
                Arguments passed directly to
                :func:`mean_sky_coordinates` for the determination of
                the sky coordinates, if they aren't provided directly.

        Returns:
            :obj:`tuple`: Two `numpy.ndarray`_ objects are returned.
            The first has the unique (non-negative) bin indices, and
            the second provides the on-sky area of that bin.
        """
        unique_bins, bin_count = numpy.unique(bin_indx, return_counts=True)
        indx = unique_bins > -1
        nbin = bin_count[indx]
        good_bins = unique_bins[indx]

        try:
            # Test that shapely was imported
            cascaded_union
        except:
            # It wasn't, so do the stupid calculation
            warnings.warn('Could not use \'shapely\' package to compute overlapping fiber area.' \
                          'Return the total fiber area.', ImportWarning)
            return good_bins, numpy.array([numpy.sum(self.area[bin_indx == b]) for b in good_bins])

        if x is None or y is None:
            x, y = self.mean_sky_coordinates(**kwargs)
        if x.size != self.nspec or y.size != self.nspec:
            raise ValueError('Must provide an x and y coordinate for each spectrum.')

        # Assume apertures are circular
        radius = numpy.sqrt(self.area/numpy.pi)
        bin_area = numpy.empty(nbin.size, dtype=float)
        for i, b in enumerate(good_bins):
            _x = x[bin_indx == b]
            _y = y[bin_indx == b]
            bin_area[i] = cascaded_union([Point(xx,yy).buffer(r,64) \
                                                    for xx, yy, r in zip(_x,_y,radius)]).area
        return good_bins, bin_area

    def central_wavelength(self, waverange=None, response_func=None, per_pixel=True, flag=None,
                           fluxwgt=False):
        """
        Determine the mean central wavelength for all spectra under
        various conditions.

        The wavelength channel is set to be the center of the
        bandpass selected by ``waverange``, weighted by the response
        function and the flux (if requested/provided). The mask is
        also incorporated in these calculations. By default (i.e., no
        wavelength limits or weighting) the wavelength channel is
        just the central wavelength of the full spectral range. (Note
        that if the spectra are binned logarithmically, this isn't
        necessarily the central wavelength *channel*.)

        Args:
            waverange (array-like, optional):
                Starting and ending wavelength over which to
                calculate the statistics. Default is to use the full
                wavelength range.
            response_func (array-like, optional):
                A two-column array with the wavelength and
                transmission of a broad-band response function to use
                as a weighting function for the calculation.
            per_pixel (:obj:`bool`, optional):
                When providing a response function, base the
                calculation on per pixel measurements, instead of per
                angstrom. Set to False for a per-angstrom
                calculation.
            flag (:obj:`str`, :obj:`list`, optional):
                One or more flag names that are considered when
                deciding if a pixel should be masked. The names
                *must* be a valid bit name as defined by
                :attr:`bitmask`. If :attr:`bitmask` is None, these
                are ignored.
            fluxwgt (:obj:`bool`, optional):
                Flag to weight by the flux when determining the mean
                coordinates.

        Returns:
            :obj:`float`: The mean central wavelength of all spectra.
        """
        if waverange is None and response_func is None and not fluxwgt:
            return (self.wave[0] + self.wave[-1])/2.

        flux = self.copy_to_masked_array(waverange=waverange, flag=flag)
        dw = numpy.ones(self.nwave, dtype=float) if per_pixel \
                else angstroms_per_pixel(self.wave, log=self.log)
        _response_func = self.interpolate_to_match(response_func)

        if fluxwgt:
            norm = numpy.ma.sum(flux*_response_func[None,:]*dw[None,:], axis=1)
            cen_wave = numpy.ma.sum(flux*self.wave[None,:]*_response_func[None,:]*dw[None,:],
                                    axis=1) / norm
            return numpy.mean(cen_wave)

        norm = numpy.ma.sum(numpy.invert(numpy.ma.getmaskarray(flux))
                            * _response_func[None,:] * dw[None,:], axis=1)
        cen_wave = numpy.ma.sum(numpy.invert(numpy.ma.getmaskarray(flux)) * self.wave[None,:]
                                * _response_func[None,:] * dw[None,:], axis=1) / norm
        return numpy.mean(cen_wave)

    # TODO: This is virtually identical to the function in DataCube.
    # Consolidate?
    def flux_stats(self, waverange=None, response_func=None, per_pixel=True, flag=None):
        r"""
        Compute the mean flux, propagated error in the mean flux, and
        mean S/N over the specified wavelength range.

        If the wavelength range is not specified, the quantities are
        calculated over the full spectral range.

        Args:
            waverange (array-like, optional):
                Starting and ending wavelength over which to
                calculate the statistics. Default is to use the full
                wavelength range.
            response_func (array-like, optional):
                A two-column array with the wavelength and
                transmission of a broad-band response function to use
                as a weighting function for the calculation.
            per_pixel (:obj:`bool`, optional):
                When providing a response function, continue to
                calculate the statistics per pixel. Set to False for
                a per-angstrom calculation.
            flag (:obj:`str`, :obj:`list`, optional):
                One or more flag names that are considered when
                deciding if a pixel should be masked. The names
                *must* be a valid bit name as defined by
                :attr:`bitmask`.

        Returns:
            `numpy.ndarray`_: Three objects are returned: the mean
            flux, the propagated variance in the mean flux, and the
            mean S/N.
    
        Raises:
            ValueError:
                Raised of a provided wavelength range object does not
                have two elements or if a response function is
                provided and has an incorrect shape.
        """
        if waverange is not None and len(waverange) != 2:
            raise ValueError('Provided wavelength range must be a two-element vector.')
        if response_func is not None:
            if len(response_func.shape) != 2:
                raise ValueError('Response function object must be two dimensional.')
            if response_func.shape[1] != 2:
                raise ValueError('Response function object must only have two columns.')

        # Grab the masked arrays
        flux = self.copy_to_masked_array(waverange=waverange, flag=flag)
        ivar = self.copy_to_masked_array(attr='ivar', waverange=waverange, flag=flag)
        snr = flux*numpy.ma.sqrt(ivar)

        # Set the response function
        dw = numpy.ones(self.nwave, dtype=float) if per_pixel \
                else angstroms_per_pixel(self.wave, log=self.log)
        _response_func = self.interpolate_to_match(response_func)

        # Get the moments
        response_integral = numpy.sum(numpy.invert(numpy.ma.getmaskarray(flux))
                                        * (_response_func*dw)[None,:], axis=1)
        signal = numpy.ma.divide(numpy.ma.sum(flux*(_response_func*dw)[None,:], axis=1),
                                 response_integral)
        variance = numpy.ma.divide(numpy.ma.sum(numpy.ma.power(ivar, -1.) \
                                    * (_response_func*dw)[None,:], axis=1), response_integral)
        snr = numpy.ma.divide(numpy.ma.sum(snr*(_response_func*dw)[None,:], axis=1),
                              response_integral)
        # Done
        return signal, variance, snr

    # Most of the rest of this deals with covariance!
    def _cube_dimensions(self, pixelscale=None, recenter=None, width_buffer=None, redo=False):
        """
        Determine the on-sky dimensions of the reconstructed image for
        all wavelength channels.

        When first calling this method, the pixel scale, recentering
        approach, and buffer must be provided, and they are saved to
        :attr:`pixelscale`, :attr:`recenter`, :attr:`width_buffer`.
        The spatial dimensions of the cube are calculated using the
        data in :attr:`xpos` and :attr:`ypos`, and the results are
        saved to :attr:`xs`, :attr:`ys`, :attr:`nx`, and :attr:`ny`.
        This calculation only needs to be done once per instance and
        settings for ``pixelscale``, ``recenter``, and
        ``width_buffer``. The calculation is forced to be done again
        if any of these are provided to the method and different from
        the saved attributes or if ``redo`` is True.

        Args:
            pixelscale (:obj:`float`, optional):
                Desired pixel scale in arcsec.  
            recenter (:obj:`bool`, optional):
                Flag to recenter the coordinate system.
            width_buffer (:obj:`int`, optional):
                Number of pixels to use as buffer for the image
                reconstruction.
            redo (:obj:`bool`, optional):
                Force the recalculation of the cube dimensions if
                they are already defined.
        """
        if self.xpos is None or self.ypos is None:
            raise ValueError('Cannot construct cube dimensions without the wavelength dependent '
                             'aperture positions.')
        # Check the input
        if self.pixelscale is None and pixelscale is None:
            raise ValueError('Must provide pixelscale on first use of _cube_dimensions.')
        if self.recenter is None and recenter is None:
            raise ValueError('Must provide recenter on first use of _cube_dimensions.')
        if self.width_buffer is None and width_buffer is None:
            raise ValueError('Must provide width_buffer on first use of _cube_dimensions.')

        # Check if the calculation needs to redone. This will always be
        # true on the first use of the method.
        _redo = redo or (pixelscale is not None and self.pixelscale != pixelscale) \
                        or (recenter is not None and self.recenter != recenter) \
                        or (width_buffer is not None and self.width_buffer != width_buffer) \
                        or self.xs is None or self.nx is None or self.ys is None or self.ny is None
        if not _redo:
            return

        # Save the parameters used to create the dimensions. Only save
        # the ones that are actually provided.
        if pixelscale is not None:
            self.pixelscale = pixelscale
        if recenter is not None:
            self.recenter = recenter
        if width_buffer is not None:
            self.width_buffer = width_buffer

        # Get the size in each dimension
        # TODO: This negative here is just to ensure that the
        # calculation matches what's done by the DRP, which defines
        # xpos as increasing from E to W.
        minx = numpy.amin(-self.xpos)
        maxx = numpy.amax(-self.xpos)
        Dx = numpy.floor(maxx-minx)

        miny = numpy.amin(self.ypos)
        maxy = numpy.amax(self.ypos)
        Dy = numpy.floor(maxy-miny)

        # Force the size to be even and the same in both dimensions
        Dx = Dx if Dx > Dy else Dy
        self.nx = int(numpy.floor(Dx/self.pixelscale)+self.width_buffer)
        if self.nx % 2 != 0:
            self.nx += 1
        self.ny = self.nx

        # Set the starting coordinate
        self.xs = -self.nx*self.pixelscale/2.
        self.ys = -self.ny*self.pixelscale/2.

        # Offset to the center, if requested
        if self.recenter:
            self.xs = self.xs + (minx+maxx)/2.0
            self.ys = self.ys + (miny+maxy)/2.0

    def _reinit_rectification_parameters(self, channel=None, pixelscale=None, rlim=None,
                                         sigma=None, recenter=None, width_buffer=None):
        """
        Determine if the rectification paraemters need to be reinitialized.
        """
        if None in [self.pixelscale, self.recenter, self.width_buffer, self.xs, self.nx, self.ys,
                    self.ny, self.rect_rlim, self.rect_sigma, self.rect_channel]:
            return True
        if channel is not None and self.rect_channel != channel:
            return True
        if pixelscale is not None and self.pixelscale != pixelscale:
            return True
        if rlim is not None and self.rect_rlim != rlim:
            return True
        if sigma is not None and self.rect_sigma != sigma:
            return True
        if recenter is not None and self.recenter != recenter:
            return True
        if width_buffer is not None and self.width_buffer != width_buffer:
            return True
        return False

    def _init_rectification_parameters(self, channel=None, pixelscale=None, rlim=None, sigma=None,
                                       recenter=None, width_buffer=None):
        """
        Perform common preparation of rectification parameters.

        Args:
            channel (:obj:`int`, optional):
                Index of the spectral channel for which to calculate
                the transfer matrix.
            pixelscale (:obj:`float`, optional):
                Desired pixel (spaxel) scale in arcsec.
            rlim (:obj:`float`, optional):
                The limiting radius of the image reconstruction
                (Gaussian) kernel in arcseconds.
            sigma (:obj:`float`, optional):
                The sigma of the image reconstruction (Gaussian)
                kernel in arcseconds.
            recenter (:obj:`bool`, optional):
                Flag to recenter the coordinate system.
            width_buffer (:obj:`int`, optional):
                Number of pixels to use as buffer for the image
                reconstruction
        """
        if not self._reinit_rectification_parameters(channel=channel, pixelscale=pixelscale,
                                                     rlim=rlim, sigma=sigma, recenter=recenter,
                                                     width_buffer=width_buffer):
            # Nothing needs to be reinitialized
            return

        # Get the cube dimensions; may not necessarily match DRP calculation
        self._cube_dimensions(pixelscale=pixelscale, recenter=recenter, width_buffer=width_buffer)

        # Set the kernel parameters if they've changed
        if rlim is not None:
            self.rect_rlim = rlim
        if sigma is not None:
            self.rect_sigma = sigma

        # Set the channel if it's changed
        if channel is not None:
            self.rect_channel = channel

    # TODO: Allow for spectrum-dependent (rlim,sigma), or scale a
    # single (rlim,sigma) by the sqrt of the area to account for
    # different size apertures?
    def rectification_transfer_matrix(self, channel, pixelscale, rlim, sigma, recenter=False,
                                      width_buffer=10, quiet=False, rej_flag=None):
        r"""
        Calculate the transfer matrix used to produce a reconstructed
        image of the fiber data at the specified wavelength channel
        using Shepard's method.

        This is done directly using the available on-sky x and y
        coordinates of each fiber as a function of wavelength taken
        from :attr:`xpos` and :attr:`ypos` attributes.

        .. todo::

            - Give more detail on the pixels at which the radius is
              calculated.
            - Describe the algorithm in more depth.

        Args:
            channel (:obj:`int`):
                Index of the spectral channel for which to calculate
                the transfer matrix.
            pixelscale (:obj:`float`):
                Desired pixel (spaxel) scale in arcsec.
            rlim (:obj:`float`):
                The limiting radius of the image reconstruction
                (Gaussian) kernel in arcseconds.
            sigma (:obj:`float`):
                The sigma of the image reconstruction (Gaussian)
                kernel in arcseconds.
            recenter (:obj:`bool`, optional):
                Flag to recenter the coordinate system.
            width_buffer (:obj:`int`, optional):
                Number of pixels to use as buffer for the image
                reconstruction
            quiet (:obj:`bool`, optional):
                Suppress terminal output
            rej_flag (:obj:`str`, optional):
                Bit name to ignore in rectification. Ignored if
                :attr:`bitmask` is None. If None, bitmasks are
                ignored. Use ``'any'`` to mask pixels with any
                flagged bits.

        Returns:
            `scipy.sparse.csr_matrix`_: Transfer matrix
            :math:`{\mathbf T}`

        """
        # Is the matrix already available?
        if self.rect_T is not None \
                and not self._reinit_rectification_parameters(channel=channel,
                                                              pixelscale=pixelscale, rlim=rlim,
                                                              sigma=sigma, recenter=recenter,
                                                              width_buffer=width_buffer):
            return self.rect_T

        # Create it!  
        self._init_rectification_parameters(channel=channel, pixelscale=pixelscale, rlim=rlim,
                                            sigma=sigma, recenter=recenter,
                                            width_buffer=width_buffer)

        # Dimensions of the sparse matrix are
        nim = self.nx*self.ny                   # The number of image pixels
        # by the number of fiber spectra (self.nspec)

        # Get the list of non-zero pixel values in the transfer matrix
        i = numpy.arange(self.nx)
        j = numpy.arange(self.ny)
        ii,jj = numpy.meshgrid(i, j, indexing='ij')         # Mesh of i,j pixel indices

        sp = numpy.empty((self.nx,self.ny), dtype=float)    # Holds spectrum index
        ij = (ii*self.ny+jj)                                # Holds image pixel index
        r2 = numpy.empty((self.nx,self.ny), dtype=float)    # Holds radii
        tot = numpy.zeros((self.nx,self.ny), dtype=float)   # Holds the sum of the weights

        s2 = numpy.square(self.rect_sigma/self.pixelscale)  # sigma^2 of Gaussian
        rl2 = numpy.square(self.rect_rlim/self.pixelscale)  # radius^2 of Gaussian limit

        non_zero_spc = numpy.empty(0, dtype=int)            # Holds triplet spectrum index
        non_zero_pix = numpy.empty(0, dtype=int)            # Holds triplet image index
        non_zero_wgt = numpy.empty(0, dtype=float)          # Holds triplet weight

        # Do not include any pixels with zero inverse variance or
        # pixels that have been masked (either by the boolean mask
        # attribute or flagged with the provided mask bits)
        mask = numpy.invert(self.ivar[:,self.rect_channel] > 0.0)
        if rej_flag is not None and self.bitmask is not None:
            _rej_flag = rej_flag if isinstance(rej_flag, list) or rej_flag != 'any' else None
            mask |= self.bitmask.flagged(self.mask[:,self.rect_channel], flag=_rej_flag)
        elif self.mask is not None:
            mask |= self.mask[:,self.rect_channel]

        # TODO: Can optimize this further?
        for k in range(self.nspec):
            if mask[k]:
                continue

            # Fill spectrum index
            sp.fill(k)

            # NOTE: Calculating full matrix is actually faster than
            # determining submatrix for calculation

            # NOTE: The negative sign in xpos is again to match the
            # calculation done by the DRP, which defines xpos as
            # increasing from E to W.

            # Calculate the distance.
            # ---- WITH RESPECT TO THE EDGE OF THE FIRST PIXEL ----
            #  - matches DRP, but why?
            r2 = numpy.square((-self.xpos[k,self.rect_channel]-self.xs)/self.pixelscale - ii) \
                 + numpy.square((self.ypos[k,self.rect_channel]-self.ys)/self.pixelscale - jj)
            # ---- WITH RESPECT TO THE CENTER OF THE FIRST PIXEL ----
#           r2 = numpy.square( (self.hdu['XPOS'].data[k,channel]-self.xs)/pixelscale-0.5 - ii) \
#                + numpy.square((self.hdu['YPOS'].data[k,channel]-self.ys)/pixelscale-0.5 - jj)

            # Append new indices and weights within rlim
            # TODO: Can this be done quicker if I'm not appending things?
            non_zero_spc = numpy.append(non_zero_spc, sp[r2 < rl2])
            non_zero_pix = numpy.append(non_zero_pix, ij[r2 < rl2])
            wgt = numpy.exp(-r2[r2 < rl2]/s2/2.0)
            tot[r2<rl2] += self.area[k] * wgt
            non_zero_wgt = numpy.append(non_zero_wgt, wgt)

            if not quiet:
                print('Transfer Matrix {:2.1%}'.format((k+1)/self.nspec), end='\r')
        if not quiet:
            print('Transfer Matrix {:2.1%}'.format(1.))

        # Normalize the result and scale by the pixel size to ensure the
        # output cube is in units of calibrated flux density per pixel
        non_zero_wgt *= numpy.square(self.pixelscale) \
                            / tot[numpy.unravel_index(non_zero_pix.astype(int), (self.nx,self.ny))]

        # Set the transfer matrix to a sparse object
        self.rect_T = sparse.coo_matrix((non_zero_wgt, (non_zero_pix, non_zero_spc)),
                                        shape=(nim,self.nspec)).tocsr()
        # TODO: Does this need to be returned?
        return self.rect_T

    # TODO: Include mask propagation

    def rectify_wavelength_plane(self, channel, pixelscale=None, rlim=None, sigma=None,
                                 recenter=False, width_buffer=10, quiet=False, rej_flag=None,
                                 return_ivar=False, return_covar=False):
        r"""
        Return a rectified image for the specified wavelength channel
        using Shepard's method.

        First, the rectification transfer matrix, :math:`{\mathbf
        T}`, is calculated using
        :func:`rectification_transfer_matrix`. Then the wavelength
        image, :math:`{\mathbf I}` is calculated:

        .. math::
        
            {\mathbf T} \times {\mathbf F} = {\mathbf I}

        where :math:`{\mathbf F}` is the vector of fluxes in the
        selected wavelength channel for all the aperture
        measurements. The calculation of :math:`{\mathbf I}` produces
        a vector, but this is reshaped into a 2D array of shape
        :math:`(N_x, N_y)` on output.

        Args:
            channel (:obj:`int`):
                Index of the spectral channel for which to calculate
                the transfer matrix.
            pixelscale (:obj:`float`, optional):
                Desired pixel (spaxel) scale in arcsec. If None,
                :attr:`pixelscale` will be used, but fault if that
                value is also None.
            rlim (:obj:`float`, optional):
                The limiting radius of the image reconstruction
                (Gaussian) kernel in arcseconds. If None,
                :attr:`rect_rlim` will be used, but fault if that
                value is also None.
            sigma (:obj:`float`, optional):
                The sigma of the image reconstruction (Gaussian)
                kernel in arcseconds. If None, :attr:`rect_sigma`
                will be used, but fault if that value is also None.
            recenter (:obj:`bool`, optional):
                Flag to recenter the coordinate system.
            width_buffer (:obj:`int`, optional):
                Number of pixels to use as buffer for the image
                reconstruction
            quiet (:obj:`bool`, optional):
                Suppress terminal output
            rej_flag (:obj:`str`, optional):
                Bit name to ignore in rectification. Ignored if
                :attr:`bitmask` is None. If None, bitmasks are
                ignored. Use ``'any'`` to mask pixels with any
                flagged bits.
            return_ivar (:obj:`bool`, optional):
                Return the nominal inverse variance image (i.e. the
                diagonal of the covariance matrix).
            return_covar (:obj:`bool`, optional):
                Return the covariance matrix. Note that the
                covariance matrix is calculated if either
                ``return_ivar`` or ``return_covar`` are True. If
                ``return_covar`` is True, ``return_ivar`` is ignored
                (because it's automatically true).

        Returns:
            `numpy.ndarray`_, :obj:`tuple`: If both ``return_ivar``
            and ``return_covar`` are False, only the reconstructed
            image is returned. If ``return_covar`` is True, a
            :class:`mangadap.util.covariance.Covariance` object is
            also returned. If ``return_covar`` is False and
            ``return_ivar`` is True, the second returned object is a
            `numpy.ndarray`_ with the propagated inverse variance
            image.
        """
        # Set the transfer matrix (set to self.rect_T; don't need to
        # keep the returned matrix)
        self.rectification_transfer_matrix(channel, pixelscale, rlim, sigma, recenter=recenter,
                                           width_buffer=width_buffer, quiet=quiet,
                                           rej_flag=rej_flag)

        flux = self.rect_T.dot(self.flux[:,channel]).reshape(self.nx,self.ny)

        # Return the regridded data with the proper shape (nx by ny)
        if not return_covar and not return_ivar:
            return flux

        covar = self._formal_covariance_matrix(quiet=quiet)
        if return_covar:
            # Return the rectified image and the full covariance matrix
            return flux, covar

        # Only return the inverse variance, not the full covariance matrix
        return flux, numpy.ma.power(numpy.diagonal(covar.toarray()),
                                    -1).filled(0.0).reshape(self.nx,self.ny)

    def _formal_covariance_matrix(self, csr=False, quiet=False):
        r"""
        Return the formal covariance matrix.
        
        The formal covariance matrix is:

        .. math::
        
             {\mathbf C} = {\mathbf T} \times {\mathbf \Sigma} \times
             {\mathbf T}^{\rm T},

        where :math:`{\mathbf \Sigma}` is the covariance matrix for the
        row-stacked spectra for the specified wavelength channel.

        This method requires that :attr:`rect_T` not be None, it
        assumes that the current transfer matrix is correct for this
        channel, and it is largely a wrapper for
        :func:`mangadap.util.covariance.Covariance.from_matrix_multiplication`.

        Args:
            csr (:obj:`bool`, optional):
                Instead of returning a
                :class:`mangadap.util.covariance.Covariance` object,
                return the covariance matrix as a
                `scipy.sparse.csr_matrix`_ object. Primarily used by
                :func:`covariance_cube` for collating the covariance
                matrix of each wavelength channel before combining
                them into a single
                :class:`mangadap.util.covariance.Covariance` object.
            quiet (:obj:`bool`, optional):
                Suppress terminal output

        Returns:
            :class:`mangadap.util.covariance.Covariance`,
            `scipy.sparse.csr_matrix`_: The covariance matrix for the
            designated wavelength channel. The return type depends on
            ``csr``.
        """
        if self.rect_T is None:
            raise ValueError('To calculate the formal covariance matrix, must first calculate '
                             'the rectification transfer matrix.')
        var = numpy.ma.power(self.ivar[:,self.rect_channel], -1.0).filled(0.0)
        covar = Covariance.from_matrix_multiplication(self.rect_T, var)
        return covar.full() if csr else covar

    def covariance_matrix(self, channel, pixelscale=None, rlim=None, sigma=None, recenter=False,
                          width_buffer=10, csr=False, quiet=False, rej_flag=None):
        r"""
        Return the covariance matrix for the specified wavelength
        channel.

        For a regrided cube image with :math:`N_x\times N_y` pixels, the
        covariance matrix has a size that is :math:`(N_x N_y)\times (N_x
        N_y)`; however, the majority of these pixels will be zero.
        Therefore, the covariance matrix is stored as a sparse matrix
        and interfaced with using the
        :class:`mangadap.util.covariance.Covariance` object class.

        The value of the covariance matrix at pixel :math:`(i,j)` is the
        covariance between pixels :math:`(n_{x,0},n_{y,0})` and
        :math:`(n_{x,1},n_{y,1})` at the specified wavelength channel of
        the reconstructed datacube image, where

        .. math::

            n_{x,0} &= \lfloor i / N_y \rfloor \\
            n_{y,0} &= i - n_{x,0} N_y \\
            n_{x,1} &= \lfloor j / N_y \rfloor \\
            n_{y,1} &= j - n_{x,1} N_y

        and :math:`\lfloor m\rfloor` is the "floor" of :math:`m`. The
        diagonal of the covariance matrix (:math:`i=j`) provides the
        variance.

        .. warning::

            Because the covariance is calculated for a higher
            dimensional array, it's important to note how the pixels
            in the image map to the relevant covariance array
            locations. See the equations above. Also see
            :func:`mangadap.util.covariance.Covariance.transpose_raw_shape`.

        The covariance matrix provided by this method is always the
        formal covariance matrix defined as

        .. math::
        
             {\mathbf C} = {\mathbf T} \times {\mathbf \Sigma} \times
             {\mathbf T}^{\rm T},

        where :math:`{\mathbf \Sigma}` is the covariance matrix for
        the row-stacked spectra for the specified wavelength channel,
        which is taken to be a diagonal matrix with the flux
        variances along the diagonal.

        Args:
            channel (:obj:`int`):
                Index of the spectral channel for which to calculate
                the covariance matrix.
            pixelscale (:obj:`float`, optional):
                Desired pixel (spaxel) scale in arcsec. If None,
                :attr:`pixelscale` will be used, but fault if that
                value is also None.
            rlim (:obj:`float`, optional):
                The limiting radius of the image reconstruction
                (Gaussian) kernel in arcseconds. If None,
                :attr:`rect_rlim` will be used, but fault if that
                value is also None.
            sigma (:obj:`float`, optional):
                The sigma of the image reconstruction (Gaussian)
                kernel in arcseconds. If None, :attr:`rect_sigma`
                will be used, but fault if that value is also None.
            recenter (:obj:`bool`, optional):
                Flag to recenter the coordinate system.
            width_buffer (:obj:`int`, optional):
                Number of pixels to use as buffer for the image
                reconstruction
            csr (:obj:`bool`, optional):
                Instead of returning a
                :class:`mangadap.util.covariance.Covariance` object,
                return the covariance matrix as a
                `scipy.sparse.csr_matrix`_ object. Primarily used by
                :func:`covariance_cube` for collating the covariance
                matrix of each wavelength channel before combining
                them into a single
                :class:`mangadap.util.covariance.Covariance` object.
            quiet (:obj:`bool`, optional):
                Suppress terminal output
            rej_flag (:obj:`str`, optional):
                Bit name to ignore in rectification. Ignored if
                :attr:`bitmask` is None. If None, bitmasks are
                ignored. Use ``'any'`` to mask pixels with any
                flagged bits.

        Returns:
            :class:`mangadap.util.covariance.Covariance`,
            `scipy.sparse.csr_matrix`_: The covariance matrix for the
            designated wavelength channel. The return type depends on
            `csr`.
        """
        # Set the transfer matrix 
        self.rectification_transfer_matrix(channel, pixelscale, rlim, sigma, recenter=recenter,
                                           width_buffer=width_buffer, quiet=quiet,
                                           rej_flag=rej_flag)
        # Return the formal covariance matrix
        return self._formal_covariance_matrix(csr=csr, quiet=quiet)

    def covariance_cube(self, channels=None, pixelscale=None, rlim=None, sigma=None,
                        recenter=False, width_buffer=10, csr=False, quiet=False, rej_flag=None):
        """
        Return the covariance matrices for many wavelength channels.

        This is primarily a wrapper for :func:`covariance_matrix`
        that iterates through one or more wavelength channels and
        constructs a 3D :class:`mangadap.util.covariance.Covariance`
        object.
        
        Args:
            channels (:obj:`int`, array-like, optional):
                Indices of the spectral channels for which to
                calculate the covariance matrix. If None, the
                covariance matrix is calculated for *all* channels.
            pixelscale (:obj:`float`, optional):
                Desired pixel (spaxel) scale in arcsec. If None,
                :attr:`pixelscale` will be used, but fault if that
                value is also None.
            rlim (:obj:`float`, optional):
                The limiting radius of the image reconstruction
                (Gaussian) kernel in arcseconds. If None,
                :attr:`rect_rlim` will be used, but fault if that
                value is also None.
            sigma (:obj:`float`, optional):
                The sigma of the image reconstruction (Gaussian)
                kernel in arcseconds. If None, :attr:`rect_sigma`
                will be used, but fault if that value is also None.
            recenter (:obj:`bool`, optional):
                Flag to recenter the coordinate system.
            width_buffer (:obj:`int`, optional):
                Number of pixels to use as buffer for the image
                reconstruction
            csr (:obj:`bool`, optional):
                Instead of returning a
                :class:`mangadap.util.covariance.Covariance` object,
                return the covariance matrix as a
                `scipy.sparse.csr_matrix`_ object. Primarily used by
                :func:`covariance_cube` for collating the covariance
                matrix of each wavelength channel before combining
                them into a single
                :class:`mangadap.util.covariance.Covariance` object.
            quiet (:obj:`bool`, optional):
                Suppress terminal output
            rej_flag (:obj:`str`, optional):
                Bit name to ignore in rectification. Ignored if
                :attr:`bitmask` is None. If None, bitmasks are
                ignored. Use ``'any'`` to mask pixels with any
                flagged bits.

        Returns:
            :class:`mangadap.util.covariance.Covariance`,
            `numpy.ndarray`_: If ``csr`` is True, the returned object
            is an `numpy.ndarray`_ of `scipy.sparse.csr_matrix`_
            types.
        """
        # Set the channels
        _channels = numpy.arange(self.nwave) if channels is None \
                        else numpy.atleast_1d(channels)
        nc = len(_channels)

        # Build and return the covariance matrices
        CovCube = numpy.empty(nc, dtype=sparse.csr_matrix)
        for i in range(nc):
            if not quiet:
                print('Covariance Cube {0}/{1}'.format(i+1,nc), end='\r')
            CovCube[i] = self.covariance_matrix(_channels[i], pixelscale=pixelscale, rlim=rlim,
                                                sigma=sigma, recenter=recenter,
                                                width_buffer=width_buffer, csr=True, quiet=True,
                                                rej_flag=rej_flag)
        if not quiet:
            print('Covariance Cube {0}/{0}'.format(nc))
        return Covariance(CovCube, input_indx=_channels) if not csr else CovCube

    def instrumental_dispersion_plane(self, channel, dispersion_factor=None, pixelscale=None,
                                      rlim=None, sigma=None, recenter=None, width_buffer=None,
                                      quiet=False, rej_flag=None):
        r"""
        Return the instrumental dispersion for the rectified
        wavelength plane.

        Method requires that :attr:`sres` is not None.

        The method first calculates the rectification transfer
        matrix, :math:`{\mathbf T}`, using
        :func:`regrid_transfer_matrix`. The transfer matrix is then
        used to construct the rectified wavelength plane image,
        :math:`{\mathbf I}`, by computing :math:`{\mathbf T} \times
        {\mathbf F} = {\mathbf I}`, where :math:`{\mathbf F}` is the
        vector of the fiber fluxes. Under the assumption that the
        line-spread-function (LSF) is Gaussian, we determine the
        instrumental dispersion for the data in the wavelength
        channel of the reconstructed CUBE by calculating the second
        moment of the weighted sum of Gaussians of the appropriate
        dispersion. Assuming all the Gaussians have the normalized
        form:

        .. math::
            
            \mathcal{N}(x|\mu=0,\sigma) = \frac{1}{\sqrt{2\pi}\sigma}
            \exp\left(-\frac{x^2}{2\sigma^2}\right),

        the combined instrumental dispersion becomes

        .. math::
            
            \sigma_{\rm inst,j}^2 = \frac{\sum_i T_{ij}
            \sigma^2_i}{\sum_i T_{ij}},

        where :math:`T_{ij}` are the elements of the transfer matrix.
        In terms of matrix multiplications, this can be written as

        .. math::

            {\mathbf S} = \frac{ {\mathbf T} \times {\mathbf V} }{
            {\mathbf T_c} },

        where :math:`{\mathbf T_c} = {\mathbf T_c} \times {\mathbf 1}`
        is the vector with the sum :math:`\sum_j T_{ij}`,
        :math:`{\mathbf V}` is the instrumental variance for all fibers
        at the designated wavelength plane, and :math:`{\mathbf S}` is
        the variance for all the spaxels in the reconstructed wavelength
        image; the division by :math:`{\mathbf T_c}` is element-wise.

        The returned matrix is the element-wise square-root of
        :math:`{\mathbf S}`, rearranged into a 2D array of size
        :attr:`nx` by :attr:`ny`.

        Args:

            channel (:obj:`int`):
                Index of the spectral channel for which to calculate
                the covariance matrix.
            dispersion_factor (:obj:`float`, optional):
                Artificially change the resolution by multiplying the
                instrumental LSF by this factor before calculating
                the reconstructed dispersion. If None, factor is
                unity.
            pixelscale (:obj:`float`, optional):
                Desired pixel (spaxel) scale in arcsec. If None,
                :attr:`pixelscale` will be used, but fault if that
                value is also None.
            rlim (:obj:`float`, optional):
                The limiting radius of the image reconstruction
                (Gaussian) kernel in arcseconds. If None,
                :attr:`rect_rlim` will be used, but fault if that
                value is also None.
            sigma (:obj:`float`, optional):
                The sigma of the image reconstruction (Gaussian)
                kernel in arcseconds. If None, :attr:`rect_sigma`
                will be used, but fault if that value is also None.
            recenter (:obj:`bool`, optional):
                Flag to recenter the coordinate system.
            width_buffer (:obj:`int`, optional):
                Number of pixels to use as buffer for the image
                reconstruction
            quiet (:obj:`bool`, optional):
                Suppress terminal output
            rej_flag (:obj:`str`, optional):
                Bit name to ignore in rectification. Ignored if
                :attr:`bitmask` is None. If None, bitmasks are
                ignored. Use ``'any'`` to mask pixels with any
                flagged bits.

        Returns:
            `numpy.ndarray`_: The reconstructed LSF dispersion for
            the specified wavelength channel.
        """
        if self.sres is None:
            raise ValueError('Spectral resolution must be available for this method.')

        # Get the instrumental dispersion
        disp = self.wave[channel]/self.sres[:,channel]/DAPConstants.sig2fwhm
        if dispersion_factor is not None:
            disp *= dispersion_factor

        # Set the transfer matrix 
        self.rectification_transfer_matrix(channel, pixelscale, rlim, sigma, recenter=recenter,
                                           width_buffer=width_buffer, quiet=quiet,
                                           rej_flag=rej_flag)

        # Rectify the data
        Tc = self.rect_T.sum(axis=1).flatten()
        Tc[numpy.invert(Tc>0)] = 1.0                # Control for zeros
        # NOTE: This returns a numpy.ndarray instead of a numpy.matrix
        return numpy.asarray(numpy.sqrt(self.rect_T.dot(numpy.square(disp)) 
                                / Tc).reshape(self.nx, self.ny))



"""
A class hierarchy that fits the emission lines.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import inspect
from pathlib import Path
import logging
import warnings

from IPython import embed

import numpy

from astropy.io import fits
import astropy.constants

from ..par.parset import KeywordParSet, ParSet

from ..util.fitsutil import DAPFitsUtil
from ..util.dapbitmask import DAPBitMask
from ..util.log import log_output
from .spatiallybinnedspectra import SpatiallyBinnedSpectra
from .stellarcontinuummodel import StellarContinuumModel
#from .elric import Elric, ElricPar
from .sasuke import Sasuke, SasukePar
from .util import replace_with_data_from_nearest_coo
from .spectralfitting import EmissionLineFit


class EmissionLineModelDef(KeywordParSet):
    """
    A class that holds the parameters necessary to perform the
    emission-line profile fits.

    The defined parameters are:

    .. include:: ../tables/emissionlinemodeldef.rst
    """
    def __init__(self, key='EFITMPL11SSPDB', minimum_snr=0.0, mom_vel_name='Ha-6564',
                 mom_disp_name=None, fitpar=None, fitclass=None, fitfunc=None, overwrite=False):

        # Use the signature to get the parameters and the default values
        sig = inspect.signature(self.__class__)
        pars = list(sig.parameters.keys())
        defaults = [sig.parameters[key].default for key in pars]

        # Remaining definitions done by hand
        in_fl = [int, float]
        par_opt = [ParSet, dict]

        values = [key, minimum_snr, mom_vel_name, mom_disp_name, fitpar, fitclass, fitfunc,
                  overwrite]
        dtypes = [str, in_fl, str, str, par_opt, None, None, bool]
        can_call = [False, False, False, False, False, False, True, False]

        descr = ['Keyword used to distinguish between different emission-line moment databases.',
                 'Minimum S/N of spectrum to fit',
                 'Name of the emission-line moments band used to set the initial velocity guess ' \
                    'for each spaxel.',
                 'Name of the emission-line moments band used to set the initial velocity ' \
                    'dispersion guess for each spaxel.',
                 'Fitting function parameters',
                 'Class used to perform the fit.',
                 'Function or method that performs the fit; **must** be callable.',
                 'If the output file already exists, redo all the calculations and overwrite it.']

        super().__init__(pars, values=values, defaults=defaults, dtypes=dtypes, can_call=can_call,
                         descr=descr)
        self._validate()

    def _validate(self):
        """
        Validate the modeling method.
        """
        if self['fitfunc'] is None:
            warnings.warn('No fitting function defined.  Defaulting to use mangadap.proc.ppxffit.')
            self['fitpar'] = SasukePar()
            self['fitclass'] = Sasuke(EmissionLineModelBitMask())
            self['fitfunc'] = self['fitclass'].fit_SpatiallyBinnedSpectra

    @classmethod
    def from_dict(cls, d):
        """
        Instantiate the object from a nested dictionary.

        The dictionary is expected to have keys for all the class keywords; see
        the class instantiation method.  It should also include the
        ``'fit_method'`` keyword used to select the type of fit to perform, and
        a nested dictionary selected by the keyword ``'fit'`` with the 
        parameters required by the fitting method.  If these are not provided,
        the default fitter, :class:`~mangadap.proc.sasuke.Sasuke` and parameters
        will be used.

        .. warning::

            Currently, the only accepted value for ``'fit_method'`` is
            ``'sasuke'``, and the ``'fit'`` dictionary must provide the
            parameters defined by :class:`~mangadap.proc.sasuke.SasukePar`.

        Args:
            d (:obj:`dict`):
                Dictionary used to instantiate the class.
        """
        # Copy over the primary keywords
        _d = {}
        for key in ['key', 'minimum_snr', 'mom_vel_name', 'mom_disp_name', 'overwrite']:
            if key in d.keys():
                _d[key] = d[key]

        # Set the spatial binning method
        if 'fit_method' not in d or d['fit_method'] in ['none', None]:
            # Use the default fitter and default fitting parameters!
            _d['fitpar'] = None
            _d['fitclass'] = None
            _d['fitfunc'] = None
        elif d['fit_method'] == 'sasuke':
            # TODO: How do I pass the emission lines to Sasuke?
            _d['fitpar'] = SasukePar.from_dict(d['fit']) if 'fit' in d else SasukePar()
            _d['fitclass'] = Sasuke(EmissionLineModelBitMask())
            _d['fitfunc'] = _d['fitclass'].fit_SpatiallyBinnedSpectra
        else:
            raise ValueError('Unrecognized fitting method.  Requires direct instantiation or '
                             'code alterations.')

        # Return the instantiation
        return super().from_dict(_d)


class EmissionLineModelBitMask(DAPBitMask):
    r"""
    Derived class that specifies the mask bits for the emission-line
    modeling. The maskbits defined are:

    .. include:: ../tables/emissionlinemodelbitmask.rst
    """
    cfg_root = 'emission_line_model_bits'


class EmissionLineModel:
    r"""
    Class that holds the emission-line model fits.

    Args:
        method (:class:`~mangadap.proc.emissionlinemodel.EmissionLineModelDef`):
            Object that defines the main parameters used for the emission-line
            model fitting.
        loggers (:obj:`list`, optional):
            List of `logging.Logger`_ objects to log progress; ignored if
            quiet=True.  Logging is done using
            :func:`~mangadap.util.log.log_output`.
        quiet (:obj:`bool`, optional):
            Suppress all terminal and logging output.
    """
    def __init__(self, method, binned_spectra, stellar_continuum=None,
                 emission_line_moments=None, redshift=None, dispersion=None, minimum_error=None,
                 method_list=None, artifact_path=None, ism_line_path=None,
                 emission_line_db_path=None, output_path=None, output_file=None, hardcopy=True,
                 overwrite=None, checksum=False, loggers=None, quiet=False):

        self.loggers = None
        self.quiet = False

        # Define the database properties
        self.method = method
        if not isinstance(self.method, EmissionLineModelDef):
            raise TypeError('Method must have type EmissionLineModelDef.')

        # Setup the pixel mask
        # Mask ISM lines if a list is provided -- KHRR
        # use nsigma=2.0 to avoid masking HeI
#        self.pixelmask = SpectralPixelMask(artdb=None if self.method['artifacts'] is None else 
#                                            ArtifactDB.from_key(self.method['artifacts'],
#                                                                directory_path=artifact_path),
#                                           emldb=None if self.method['ism_mask'] is None else
#                                            EmissionLineDB.from_key(self.method['ism_mask'],
#                                                                    directory_path=ism_line_path),
#                                           waverange=self.method['waverange'], nsig=2.0)

        # Setup the emission-line database
#        self.emldb = None if self.method['emission_lines'] is None else \
#                        EmissionLineDB.from_key(self.method['emission_lines'],
#                                                directory_path=emission_line_db_path)
#        self.neml = None if self.emldb is None else self.emldb.size

#        self.emldb = None
        self.neml = None

        self.binned_spectra = None
        self.redshift = None
        self.dispersion = None
        self.stellar_continuum = None

        # Define the output directory and file
        self.directory_path = None      # Set in default_paths
        self.output_file = None
        self.hardcopy = None

        # Initialize the objects used in the assessments
        self.bitmask = EmissionLineModelBitMask()

        self.hdu = None
        self.checksum = checksum
        self.shape = None
        self.spatial_shape = None
        self.spatial_index = None
        self.spectral_arrays = None
        self._assign_spectral_arrays()
        self.image_arrays = None
        self._assign_image_arrays()
        self.nwave = None

        self.nmodels = None
        self.missing_models = None

        # Fit the binned spectra
        self.fit(binned_spectra, stellar_continuum=stellar_continuum,
                 emission_line_moments=emission_line_moments, redshift=redshift,
                 dispersion=dispersion, minimum_error=minimum_error, output_path=output_path,
                 output_file=output_file, hardcopy=hardcopy, overwrite=overwrite, loggers=loggers,
                 quiet=quiet)

    def __getitem__(self, key):
        return self.hdu[key]

    @staticmethod
    def default_paths(cube, method_key, rdxqa_method, binning_method, stelcont_method=None,
                      output_path=None, output_file=None):
        """
        Set the default directory and file name for the output file.

        Args:
            cube (:class:`mangadap.datacube.datacube.DataCube`):
                Datacube to analyze.
            method_key (:obj:`str`):
                Keyword designating the method used for the reduction
                assessments.
            rdxqa_method (:obj:`str`):
                The method key for the basic assessments of the datacube.
            binning_method (:obj:`str`):
                The method key for the spatial binning.
            stelcont_method (:obj:`str`, optional):
                The method key for the stellar-continuum fitting method.  If
                None, not included in output file name.
            output_path (:obj:`str`, `Path`_, optional):
                The path for the output file.  If None, the current working
                directory is used.
            output_file (:obj:`str`, optional):
                The name of the output "reference" file. The full path of the
                output file will be :attr:`directory_path`/:attr:`output_file`.
                If None, the default is to combine ``cube.output_root`` and the
                method keys.  The order of the keys is the order of operations
                (rdxqa, binning, stellar continuum, emission-line model).

        Returns:
            :obj:`tuple`: Returns a `Path`_ with the output directory and a
            :obj:`str` with the output file name.
        """
        directory_path = Path('.').resolve() if output_path is None \
                                else Path(output_path).resolve()
        method = f'{rdxqa_method}-{binning_method}'
        if stelcont_method is not None:
            method = f'{method}-{stelcont_method}'
        method = f'{method}-{method_key}'
        _output_file = f'{cube.output_root}-{method}.fits.gz' if output_file is None \
                            else output_file
        return directory_path, _output_file


    def _initialize_primary_header(self, hdr=None):
        """
        Construct the primary header for the reference file.

        Args:
            hdr (`astropy.io.fits.Header`_, optional):
                Input base header for added keywords. If None, uses
                the datacube header from :attr:`binned_spectra` (if
                there is one) and then cleans the header using
                :func:`mangadap.util.fitsutil.DAPFitsUtil.clean_dap_primary_header`.

        Returns:
            `astropy.io.fits.Header`_: Initialized header object.
        """
        # Copy the from the DRP and clean it
        if hdr is None:
            hdr = self.binned_spectra.cube.prihdr.copy()
            hdr = DAPFitsUtil.clean_dap_primary_header(hdr)
        
        hdr['AUTHOR'] = 'Kyle B. Westfall <westfall@ucolick.org>'
        hdr['ELFKEY'] = (self.method['key'], 'Emission-line modeling method keyword')
        hdr['ELFMINSN'] = (self.method['minimum_snr'], 'Minimum S/N of spectrum to include')
#        hdr['ARTDB'] = (self.method['artifacts'], 'Artifact database keyword')
        hdr['EMLDB'] = (self.method['fitpar']['emission_lines'], 'Emission-line database keyword')
        if self.stellar_continuum is not None:
            hdr['SCKEY'] = (self.stellar_continuum.method['key'], 'Stellar-continuum model keyword')
        hdr['NELMOD'] = (self.nmodels, 'Number of unique emission-line models')
        return hdr

    def _assign_input_kinematics(self, emission_line_moments, redshift, dispersion,
                                 default_dispersion=100.0):
        """
        Set the initial redshift and velocity dispersion for each
        spectrum.
        
        In terms of precedence, directly provided redshifts and
        dispersions override those in the EmissionLineMoments model,
        which overrides those in the StellarContinuumModel object.  The
        default_dispersion does *not* take precedence over *any*
        provided disperison.

        If emission_line_moments, redshift, and
        :attr:`stellar_continuum` are all None, the redshift is set to
        0.0.  If dispersion is also None, the dispersion is set to
        default_dispersion.

        To get the stellar kinematics, the function calls
        :func:`mangadap.proc.stellarcontinuummodel.StellarContinuumModel.matched_kinematics`.
        In this function, the provided redshift and dispersion must be a
        single value or None; therefore, the means of any vectors
        provided as redshift or disperison are passsed to this function
        instead of the full vector.

        Args:
            emission_line_moments (:class:`mangadap.proc.emissionlinemoments.EmissionLineMoments`):
                Object with the results of the emission-line-moment
                measurements
            redshift (:obj:`float`, `numpy.ndarray`_):
                Redshifts (:math:`z`) to use for each spectrum.
            dispersion (:obj:`float`, `numpy.ndarray`_):
                Velocity dispersion (km/s) to use for each spectrum.
            default_dispersion (:obj:`float`, `numpy.ndarray`_, optional):
                Default velocity dispersion to use (km/s), if
                relevant.
        """

        #---------------------------------------------------------------
        # Get the redshift and dispersion measured for the stars or
        # based on the moments if the relevant objects are provided
        obj_redshift, obj_dispersion = (None, None)
        if emission_line_moments is not None:
            names = emission_line_moments.channel_names()
            vel_channel = -1 if self.method['mom_vel_name'] is None \
                                else numpy.arange(len(names))[self.method['mom_vel_name']
                                                                == names][0]
            if vel_channel > -1:
                obj_redshift = numpy.full(self.binned_spectra.nbins, 0.0, dtype=float)
            
                mom1_masked = emission_line_moments['ELMMNTS'].data['MASK'][:,vel_channel] > 0
                # - Use the 1st moment of the named line
                obj_redshift[ emission_line_moments['ELMMNTS'].data['BINID_INDEX'] ] \
                        = emission_line_moments['ELMMNTS'].data['MOM1'][:,vel_channel] \
                                / astropy.constants.c.to('km/s').value

                # - For missing bins in the moment measurements and bad
                # line moment measurements, use the value for the
                # nearest good bin
                bad_bins = numpy.append(emission_line_moments.missing_bins,
                                        emission_line_moments['ELMMNTS'].data['BINID']\
                                            [mom1_masked]).astype(int)
                if len(bad_bins) > 0 and len(bad_bins) != self.binned_spectra.nbins:
                    nearest_good_bin_index = self.binned_spectra.find_nearest_bin(bad_bins,
                                                                                  indices=True)
                    bad_bin_index = self.binned_spectra.get_bin_indices(bad_bins)
                    obj_redshift[bad_bin_index] = obj_redshift[nearest_good_bin_index]

            disp_channel = -1 if self.method['mom_disp_name'] is None \
                                else numpy.where(self.method['mom_disp_name'] == names)[0]
            if disp_channel > -1:
                obj_dispersion = numpy.full(self.binned_spectra.nbins, default_dispersion,
                                            dtype=float)
            
                mom2_masked = emission_line_moments['ELMMNTS'].data['MASK'][:,disp_channel] > 0
                # - Use the 2nd moment of the named line
                obj_dispersion[ emission_line_moments['ELMMNTS'].data['BINID_INDEX'] ] \
                                = emission_line_moments['ELMMNTS'].data['MOM2'][:,disp_channel]

                # - For missing bins in the moment measurements and bad
                # line moment measurements, use the value for the
                # nearest good bin
                bad_bins = numpy.append(emission_line_moments.missing_bins,
                                        emission_line_moments['ELMMNTS'].data['BINID']\
                                            [mom2_masked]).astype(int)
                if len(bad_bins) > 0 and len(bad_bins) != self.binned_spectra.nbins:
                    nearest_good_bin_index = self.binned_spectra.find_nearest_bin(bad_bins,
                                                                                  indices=True)
                    bad_bin_index = self.binned_spectra.get_bin_indices(bad_bins)
                    obj_dispersion[bad_bin_index] = obj_dispersion[nearest_good_bin_index]

        if self.stellar_continuum is not None and (obj_redshift is None or obj_dispersion is None):
            sc_redshift, sc_dispersion = \
                    self.stellar_continuum.matched_kinematics(self.binned_spectra['BINID'].data,
                                redshift=None if redshift is None else numpy.ma.mean(redshift),
                                dispersion=default_dispersion if dispersion is None 
                                                                else numpy.ma.mean(dispersion),
                                corrected=True, nearest=True,
                                missing=self.binned_spectra.missing_bins)
            if obj_redshift is None:
                obj_redshift = sc_redshift
            if obj_dispersion is None:
                obj_dispersion = sc_dispersion
        #---------------------------------------------------------------

        # Redshift: use the stellar continuum values if present,
        # otherwise set the default to 0.
        self.redshift = obj_redshift if redshift is None else \
                            numpy.zeros(self.binned_spectra.nbins, dtype=float)
        if redshift is not None:
            _redshift = numpy.atleast_1d(redshift)
            if len(_redshift) not in [ 1, self.binned_spectra.nbins ]:
                raise ValueError('Provided redshift must be either a single value or match the ' \
                                 'number of binned spectra.')
            self.redshift = numpy.full(self.binned_spectra.nbins, redshift, dtype=float) \
                                if len(_redshift) == 1 else _redshift.copy()

        # Same for dispersion
        # Redshift: use the stellar continuum values if present,
        # otherwise set the default to 0.
        self.dispersion = obj_dispersion if dispersion is None else \
                            numpy.full(self.binned_spectra.nbins, default_dispersion, dtype=float)
        if dispersion is not None:
            _dispersion = numpy.atleast_1d(dispersion)

            if len(_dispersion) not in [ 1, self.binned_spectra.nbins ]:
                raise ValueError('Provided dispersion must be either a single value or match the ' \
                                 'number of binned spectra.')
            self.dispersion = numpy.full(self.binned_spectra.nbins, dispersion, dtype=float) \
                                if len(_dispersion) == 1 else _dispersion.copy()


    def _add_method_header(self, hdr, model_binid=None):
        """Add fitting method information to the header."""
        if self.method['fitclass'] is not None:
            try:
                hdr['ELTYPE'] = self.method['fitclass'].fit_type
                hdr['ELMETH'] = self.method['fitclass'].fit_method
            except:
                if not self.quiet and self.hardcopy:
                    warnings.warn('Fit class object does not have fit_type and/or fit_method ' \
                                  'attributes.  No parameters written to header.')
        if self.method['fitpar'] is not None:
            try:
                hdr = self.method['fitpar'].toheader(hdr)
            except:
                if not self.quiet and self.hardcopy:
                    warnings.warn('Fit parameter class has no toheader() function.  No ' \
                                  'parameters written to header.')

        # If the model bin IDs are provided, assume the data was
        # rebinned, meaning that the bin ID of this object is not
        # directly tied to the bin ID for the spatially binned spectra
        # object (self.binned_spectra)
        hdr['ELREBIN'] = (model_binid is not None, 'Bin IDs disconnected from SC binning')

        return hdr


    def _finalize_cube_mask(self, mask):
        """
        Finalize the mask after the 2D mask has been reconstructed
        into a 3D cube.

        This mostly handles the masks for regions outside the
        datacube field of view.

        Note that the input mask is both edited in-place and
        returned.

        .. todo::

            - This needs to be abstracted for non-DRP datacubes.
            - Describe MAPMASK usage

        Args:
            mask (`numpy.ndarray`_):
                3D array with the current bitmask data.

        Returns:
            `numpy.ndarray`_: Edited bitmask data.
        """
        # Turn on the flag stating that the pixel wasn't used
        indx = self.binned_spectra.bitmask.flagged(self.binned_spectra.cube.mask,
                                                   flag=self.binned_spectra.do_not_fit_flags())
        mask[indx] = self.bitmask.turn_on(mask[indx], 'DIDNOTUSE')

        # Turn on the flag stating that the pixel has a foreground star
        indx = self.binned_spectra.bitmask.flagged(self.binned_spectra.cube.mask, flag='FORESTAR')
        mask[indx] = self.bitmask.turn_on(mask[indx], 'FORESTAR')

        # Propagate the MAPMASK to the full cube
        for f in ['DIDNOTUSE', 'FORESTAR', 'LOW_SNR']:
            indx = self.bitmask.flagged(self['MAPMASK'].data, f)
            mask[indx,:] = self.bitmask.turn_on(mask[indx,:], f)

        return mask

    def _assign_spectral_arrays(self):
        self.spectral_arrays = [ 'FLUX', 'BASE', 'MASK' ]

    def _assign_image_arrays(self):
        """
        Set :attr:`image_arrays`, which contains the list of extensions
        in :attr:`hdu` that are on-sky image data.
        """
        self.image_arrays = [ 'BINID' ]

    def _get_missing_models(self, unique_bins=None, debug=False):
        if unique_bins is None:
            good_snr = self.binned_spectra.above_snr_limit(self.method['minimum_snr'], debug=debug)
            return numpy.sort(self.binned_spectra.missing_bins + 
                        self.binned_spectra['BINS'].data['BINID'][numpy.invert(good_snr)].tolist())
        return SpatiallyBinnedSpectra._get_missing_bins(unique_bins)

    def _get_line_fit_metrics(self, model_flux, model_base, model_mask, model_eml_par,
                              model_binid=None, window=15, debug=False):
        """
        Compute fit-quality metrics near each fitted line.

        Primarily a wrapper of :func:`EmissionLineFit.line_metrics`.

        Args:
            model_flux (`numpy.ndarray`_):
                The best-fitting emission-line model spectra.
            model_base (`numpy.ndarray`_):
                The "baseline" of the emission-line model defined as the
                difference between the best-fitting stellar-continuum
                model and the best-fitting continuum from the
                emission-line modeling procedure.
            model_mask (`numpy.ndarray`_):
                A boolean or integer (bitmask) array with the mask for
                the model data.
            model_eml_par (:class:`~mangadap.proc.spectralfitting.EmissionLineFitDataTable`):
                A record array with the best-fitting results and
                parameters for each emission line.  The format is
                expected to be given by
                :func:`EmissionLineFit._per_emission_line_dtype`.  *This
                object is edited in place and returned.*
            model_binid (`numpy.ndarray`_, optional):
                A 2D array with the bin ID assigned to each spaxel with
                a fitted emission-line model.  The number of
                non-negative bin IDs must match the number of fitted
                models.  If None, this is generated from the BINID
                column in the model_eml_par object.
            window (:obj:`int`, optional):
                The window size in pixels around each line to use for
                calculation of the figures of merit.

        Returns:
            `numpy.recarray`_: Return the input ``model_eml_par``
            after filling the LINE_PIXC, AMP, ANR, LINE_NSTAT,
            LINE_CHI2, LINE_RMS, and LINE_FRMS columns.
        """
        # Get the baseline continuum
        bin_indx = DAPFitsUtil.downselect_bins(self.binned_spectra['BINID'].data.ravel(),
                                            model_eml_par['BINID']).reshape(self.spatial_shape) \
                            if model_binid is None else model_binid
#        continuum = numpy.ma.MaskedArray(numpy.zeros(model_base.shape, dtype=float)) \
#                        if self.stellar_continuum is None else \
#                    self.stellar_continuum.fill_to_match(bin_indx, missing=self.missing_models)

        # If stellar continuum available, masked array returned by
        # fill_to_match needs to be filled with zeros to correctly
        # reconstruct the emission-line continuum.
        if self.stellar_continuum is None:
            continuum = model_base
        else:
            continuum = self.stellar_continuum.fill_to_match(bin_indx,
                                                             missing=self.missing_models).filled(0.)
            indx = numpy.unique(bin_indx.ravel())
            if indx[0] == -1:
                indx = indx[1:]
            continuum = continuum[indx,:] + model_base

        # Interpret the input mask as either a boolean or bitmask array
        if model_mask.dtype is numpy.dtype('bool'):
            _model_mask = numpy.zeros(model_flux.shape, dtype=self.bitmask.minimum_dtype())
            _model_mask[model_mask] = self.bitmask.turn_on(_model_mask[model_mask], 'DIDNOTUSE')
        else:
            _model_mask = model_mask

        # Compute the full (continuum + emission lines) model spectra
        model = numpy.ma.MaskedArray(model_flux+continuum, mask=_model_mask>0)

        pixelmask = None if 'pixelmask' not in self.method['fitpar'].keys() \
                        else self.method['fitpar']['pixelmask']

        # Get the fitted spectra
        if self.method['fitpar']['deconstruct_bins'] != 'ignore':
            # The models should be for the individual spaxels
            bins_to_fit = EmissionLineFit.select_binned_spectra_to_fit(self.binned_spectra,
                                                        minimum_snr=self.method['minimum_snr'],
                                                        stellar_continuum=self.stellar_continuum,
                                                                       debug=debug)
            # NOTE: bins_to_fit only used in this spaxel selection if debug=True
            spaxel_to_fit = EmissionLineFit.select_spaxels_to_fit(self.binned_spectra,
                                                        minimum_snr=self.method['minimum_snr'],
                                                        bins_to_fit=bins_to_fit, debug=debug)
            wave, flux, ferr, _ = EmissionLineFit.get_spectra_to_fit(self.binned_spectra,
                                                                     pixelmask=pixelmask,
                                                                     select=spaxel_to_fit,
                                                                     error=True,
                                                                     original_spaxels=True)
        else:
            # The models should be for the binned spectra
            bins_to_fit = EmissionLineFit.select_binned_spectra_to_fit(self.binned_spectra,
                                                        minimum_snr=self.method['minimum_snr'],
                                                        stellar_continuum=self.stellar_continuum,
                                                                       debug=debug)
            wave, flux, ferr, _ = EmissionLineFit.get_spectra_to_fit(self.binned_spectra,
                                                                     pixelmask=pixelmask,
                                                                     select=bins_to_fit, error=True)

        # Get the line metrics
        return EmissionLineFit.line_metrics(self.method['fitpar'].emldb, wave, flux, ferr, model,
                                            model_eml_par, bitmask=self.bitmask, window=window)

    def _construct_2d_hdu(self, good_snr, model_flux, model_base, model_mask, model_eml_par,
                          model_fit_par=None, model_binid=None):
        """
        Construct :attr:`hdu` that is held in memory for manipulation of
        the object.  See :func:`construct_3d_hdu` if you want to convert
        the object into a DRP-like datacube.
        """
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, 'Constructing hdu ...')

        # Initialize the headers
        pri_hdr = self._initialize_primary_header()
        pri_hdr = self._add_method_header(pri_hdr, model_binid=model_binid)
        map_hdr = DAPFitsUtil.build_map_header(self.binned_spectra.cube.fluxhdr,
                                               'K Westfall <westfall@ucolick.org>')

        # Get the spatial map mask
        map_mask = numpy.zeros(self.spatial_shape, dtype=self.bitmask.minimum_dtype())

        # Always isolate any spaxels with foreground stars
        indx = self.binned_spectra.bitmask.flagged(self.binned_spectra['MAPMASK'].data, 'FORESTAR')
        map_mask[indx] = self.bitmask.turn_on(map_mask[indx], 'FORESTAR')

        # Allow the fitting function to redefine the bin IDs associated
        # with each model
        if model_binid is None:
            # Add any spaxel not used because it was flagged by the
            # binning step
            indx = self.binned_spectra['MAPMASK'].data > 0
            map_mask[indx] = self.bitmask.turn_on(map_mask[indx], 'DIDNOTUSE')

            # Get the bins that were below the S/N limit
            indx = numpy.invert(DAPFitsUtil.reconstruct_map(self.spatial_shape,
                                                    self.binned_spectra['BINID'].data.ravel(),
                                                    good_snr, dtype='bool')) & (map_mask == 0)
            map_mask[indx] = self.bitmask.turn_on(map_mask[indx], 'LOW_SNR')

            # Get the bin ids with fitted models
            bin_indx = DAPFitsUtil.downselect_bins(self.binned_spectra['BINID'].data.ravel(),
                                            model_eml_par['BINID']).reshape(self.spatial_shape)
        else:
            # Assume any model with a binid less than zero is from a
            # spaxel that was not used
            indx = model_binid < 0
            map_mask[indx] = self.bitmask.turn_on(map_mask[indx], 'DIDNOTUSE')

            # The number of valid bins MUST match the number of
            # measurements
            nvalid = numpy.sum(numpy.invert(indx))
            if nvalid != len(model_eml_par):
                raise ValueError('Provided model id does not match the number of measurements.')

            # Get the bin ids with fitted models
            bin_indx = model_binid

        # Allow the fitting function to return a boolean model mask
        if model_mask.dtype is numpy.dtype('bool'):
            _model_mask = numpy.zeros(model_flux.shape, dtype=self.bitmask.minimum_dtype())
            _model_mask[model_mask] = self.bitmask.turn_on(_model_mask[model_mask], 'DIDNOTUSE')
        else:
            _model_mask = model_mask

        # Save the data to the hdu attribute
        par_hdu = fits.BinTableHDU(data=None, name='PAR') if self.method['fitpar'].emldb is None \
                        else self.method['fitpar'].emldb.to_datatable().to_hdu(name='PAR')
        fit_hdu = fits.BinTableHDU(data=None, name='FIT') if model_fit_par is None else \
                        model_fit_par.to_hdu(name='FIT', add_dim=True)

        self.hdu = fits.HDUList([fits.PrimaryHDU(header=pri_hdr),
                                 fits.ImageHDU(data=model_flux.data, name='FLUX'),
                                 fits.ImageHDU(data=model_base.data, name='BASE'),
                                 fits.ImageHDU(data=_model_mask, name='MASK'),
                                 self.binned_spectra['WAVE'].copy(),
                                 fits.ImageHDU(data=bin_indx, header=map_hdr, name='BINID'),
                                 fits.ImageHDU(data=map_mask, header=map_hdr, name='MAPMASK'),
                                 par_hdu, fit_hdu,
                                 model_eml_par.to_hdu(name='EMLDATA', add_dim=True)
                                ])

    def file_name(self):
        """Return the name of the output file."""
        return self.output_file

    def file_path(self):
        """Return the full path to the output file."""
        if self.directory_path is None or self.output_file is None:
            return None
        return self.directory_path / self.output_file


    def fit(self, binned_spectra, stellar_continuum=None, emission_line_moments=None,
            redshift=None, dispersion=None, minimum_error=None, output_path=None, output_file=None,
            hardcopy=True, overwrite=None, loggers=None, quiet=False):
        """
        Fit the emission lines.
        """
        # Initialize the reporting
        if loggers is not None:
            self.loggers = loggers
        self.quiet = quiet

        # SpatiallyBinnedSpectra object always needed
        if binned_spectra is None:
            raise ValueError('Must provide spectra object for fitting.')
        if not isinstance(binned_spectra, SpatiallyBinnedSpectra):
            raise TypeError('Must provide a valid SpatiallyBinnedSpectra object!')
        if binned_spectra.hdu is None:
            raise ValueError('Provided SpatiallyBinnedSpectra object is undefined!')
        self.binned_spectra = binned_spectra

        # Continuum accounts for underlying absorption
        if stellar_continuum is not None \
                and not isinstance(stellar_continuum, StellarContinuumModel):
            raise TypeError('Must provide a valid StellarContinuumModel object.')
        self.stellar_continuum = stellar_continuum

        # Emission-line moments can be used to set the initial guess
        # kinematics
        if emission_line_moments is not None:
            # No longer checks for this; errors will be raised
            # just by trying to use the object...
#            if not isinstance(emission_line_moments, EmissionLineMoments):
#                raise TypeError('Must provide a valid EmissionLineMoments object.')
            names = emission_line_moments.channel_names()
            if self.method['mom_vel_name'] is not None \
                        and self.method['mom_vel_name'] not in names:
                raise ValueError('{0}'.format(self.method['mom_vel_name']) +
                                 ' is not a valid name in the EmissionLineMoments object.')
            if self.method['mom_disp_name'] is not None \
                        and self.method['mom_disp_name'] not in names:
                raise ValueError('{0}'.format(self.method['mom_disp_name']) +
                                 ' is not a valid name in the EmissionLineMoments object.')

        self.shape = self.binned_spectra.shape
        self.spatial_shape =self.binned_spectra.spatial_shape
        self.spatial_index = self.binned_spectra.spatial_index.copy()
        self.nwave = self.binned_spectra.nwave

        # Get the guess kinematics
        self._assign_input_kinematics(emission_line_moments, redshift, dispersion)

        #---------------------------------------------------------------
        # Get the good spectra
#        debug=True
        debug=False
#        good_snr = EmissionLineFit.select_binned_spectra_to_fit(self.binned_spectra,
#                                                        minimum_snr=self.method['minimum_snr'],
#                                                        stellar_continuum=self.stellar_continuum,
#                                                        debug=debug)
        good_bins = EmissionLineFit.select_binned_spectra_to_fit(self.binned_spectra,
                                                        minimum_snr=self.method['minimum_snr'],
                                                        stellar_continuum=self.stellar_continuum,
                                                        debug=debug)
        # NOTE: bins_to_fit only used in this spaxel selection if debug=True
        good_spax = EmissionLineFit.select_spaxels_to_fit(self.binned_spectra,
                                                        minimum_snr=self.method['minimum_snr'],
                                                        bins_to_fit=good_bins, debug=debug)

        # Report
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, '-'*50)
            log_output(loggers, 1, logging.INFO, '{0:^50}'.format('EMISSION-LINE PROFILE FITTING'))
            log_output(self.loggers, 1, logging.INFO, '-'*50)
            log_output(self.loggers, 1, logging.INFO, 'Number of binned spectra: {0}'.format(
                                                            self.binned_spectra.nbins))
            if len(self.binned_spectra.missing_bins) > 0:
                log_output(self.loggers, 1, logging.INFO, 'Missing bins: {0}'.format(
                                                            len(self.binned_spectra.missing_bins)))
            log_output(self.loggers, 1, logging.INFO, 'Bins with good S/N: {0}'.format(
                                                            numpy.sum(good_bins)))
            log_output(self.loggers, 1, logging.INFO, 'Spaxels with good S/N: {0}'.format(
                                                            numpy.sum(good_spax)))
#            log_output(self.loggers, 1, logging.INFO, 'Bin deconstruction method: {0}'.format(
#                                                            self.method['deconstruct_bins']))
            
        if numpy.sum(good_bins) == 0:
            raise ValueError('No good spectra to fit!')

        #---------------------------------------------------------------
        # Fill in any remaining binning parameters
#        self._fill_method_par()

        if self.method['fitpar'] is not None and hasattr(self.method['fitpar'], 'fill'):
            self.method['fitpar'].fill(self.binned_spectra,
                                       stellar_continuum=self.stellar_continuum,
                                       guess_vel=self.redshift*astropy.constants.c.to('km/s').value,
                                       guess_sig=self.dispersion, output_path=self.directory_path,
                                       #hardcopy=self.tpl_hardcopy, symlink_dir=self.symlink_dir,
                                       loggers=self.loggers, quiet=self.quiet)

        _overwrite = self.method['overwrite'] if overwrite is None else overwrite

        # (Re)Set the output paths
        self.directory_path, self.output_file \
                = EmissionLineModel.default_paths(self.binned_spectra.cube, self.method['key'],
                        self.binned_spectra.rdxqa.method['key'], self.binned_spectra.method['key'],
                        stelcont_method=None if self.stellar_continuum is None
                                        else self.stellar_continuum.method['key'],
                        output_path=output_path, output_file=output_file)

        #---------------------------------------------------------------
        # Check that the file path is defined
        ofile = self.file_path()
        if ofile is None:
            raise ValueError('File path for output file is undefined!')

        # Report
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, f'Output path: {self.directory_path}')
            log_output(self.loggers, 1, logging.INFO, f'Output file: {self.output_file}')

        #---------------------------------------------------------------
        # If the file already exists, and not overwriting, just read the
        # file
        if ofile.exists() and not _overwrite:
            self.hardcopy = True
            if not self.quiet:
                log_output(self.loggers, 1, logging.INFO, 'Using existing file')
            self.read(checksum=self.checksum, debug=debug)
            if not self.quiet:
                log_output(self.loggers, 1, logging.INFO, '-'*50)
            return

        #---------------------------------------------------------------
        # Fit the spectra
        # Mask should be fully defined within the fitting function This
        # should also add the equivalent widths.  TODO: Shouldn't the
        # equivalent widths be done here?
        model_flux, model_base, model_mask, model_fit_par, model_eml_par, model_binid = \
                self.method['fitfunc'](self.binned_spectra,
                                       stellar_continuum=self.stellar_continuum,
                                       good_bins=good_bins, good_spax=good_spax,
                                       par=self.method['fitpar'], loggers=self.loggers,
                                       quiet=self.quiet, debug=debug)

        # Impose a minimum error because of the conversion to inverse
        # variance when constructing a MAPS file; applied to the error
        # quantities from spectralfitting.EmissionLineFit dtype
        # parameters: FLUXERR, KINERR, EWERR
        # TODO: Add a maskbit as well?  Yes, but in MAPS construction
        indx = model_eml_par['FLUXERR'] < (numpy.finfo(model_eml_par['FLUX'].dtype).eps \
                                            if minimum_error is None else minimum_error)
        model_eml_par['FLUXERR'][indx] = 0.0
        indx = model_eml_par['KINERR'] < (numpy.finfo(model_eml_par['KIN'].dtype).eps \
                                            if minimum_error is None else minimum_error)
        model_eml_par['KINERR'][indx] = 0.0
        indx = model_eml_par['EWERR'] < (numpy.finfo(model_eml_par['EW'].dtype).eps \
                                            if minimum_error is None else minimum_error)
        model_eml_par['EWERR'][indx] = 0.0

        # TODO: Test if, when deconstruct_bins is true, the model_binid
        # spaxels (with fits) all have unique bin IDs.

        # The emission-line fitting can proceed without knowing which
        # lines are fit; this sets the number of lines
        if self.neml is None:
            self.neml = model_eml_par['FLUX'].shape[1]

        # The number of models returned should be the number of "good" binned
        # spectra if the fitter does *not* deconstruct the bins.
        # TODO: This may now cause problems with Elric...
        if self.method['fitpar']['deconstruct_bins'] == 'ignore' \
                and model_flux.shape[0] != numpy.sum(good_bins):
            print(model_flux.shape[0])
            print(numpy.sum(good_bins))
            raise ValueError('Unexpected returned shape of fitted emission-line models.')
       
        #---------------------------------------------------------------
        # Set the number of models and the missing models
        self.nmodels = model_flux.shape[0]
        unique_bins = None if model_binid is None else numpy.unique(model_binid.ravel())
        self.missing_models = self._get_missing_models(unique_bins=unique_bins, debug=debug)

        # Report
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, f'Fitted models: {self.nmodels}')
            if len(self.missing_models) > 0:
                log_output(self.loggers, 1, logging.INFO,
                           f'Missing models: {len(self.missing_models)}')
            log_output(self.loggers, 1, logging.INFO, 'Calculating fit metrics')

        # Compute the line-fit metrics
        model_eml_par = self._get_line_fit_metrics(model_flux, model_base, model_mask,
                                                   model_eml_par, model_binid=model_binid,
                                                   debug=debug)

        # Construct the 2d hdu list that only contains the fitted models
        self.hardcopy = hardcopy
        self._construct_2d_hdu(good_bins, model_flux, model_base, model_mask, model_eml_par,
                               model_fit_par=model_fit_par, model_binid=model_binid)

        #---------------------------------------------------------------
        # Write the data, if requested
        if self.hardcopy:
            if not self.directory_path.exists():
                self.directory_path.mkdir(parents=True)
            self.write(overwrite=_overwrite)
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, '-'*50)


    def construct_3d_hdu(self):
        """
        Reformat the model spectra into a cube matching the shape of
        the DRP fits file.
        """
        # Report
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO,
                       'Constructing emission-line model datacube ...')

        bin_indx = self.hdu['BINID'].data.copy()
        model_flux = self.hdu['FLUX'].data.copy()
        model_base = self.hdu['BASE'].data.copy()
        model_mask = self.hdu['MASK'].data.copy()

        flux, base, mask = DAPFitsUtil.reconstruct_cube(self.shape, bin_indx.ravel(),
                                                        [model_flux, model_base, model_mask])

        mask = self._finalize_cube_mask(mask)

        # Primary header is identical regardless of the shape of the
        # extensions
        hdr = self.hdu['PRIMARY'].header.copy()
        cube_hdr = DAPFitsUtil.build_cube_header(self.binned_spectra.cube,
                                                 'K Westfall <westfall@ucolick.org>')

        par_hdu = self.hdu['PAR'].copy() if 'data' in self.hdu['PAR'].__dict__ \
                        else fits.BinTableHDU(data=None, name='PAR')
        fit_hdu = self.hdu['FIT'].copy() if 'data' in self.hdu['FIT'].__dict__ \
                        else fits.BinTableHDU(data=None, name='FIT')

        # Return the converted hdu without altering the internal hdu
        return fits.HDUList([ fits.PrimaryHDU(header=hdr),
                              fits.ImageHDU(data=flux, header=cube_hdr, name='FLUX'),
                              fits.ImageHDU(data=base, header=cube_hdr, name='BASE'),
                              fits.ImageHDU(data=mask, header=cube_hdr, name='MASK'),
                              self.hdu['WAVE'].copy(),
                              self.hdu['BINID'].copy(),
                              par_hdu,
                              fit_hdu,
                              self.hdu['EMLDATA'].copy()
                            ])


    # Exact same function as used by SpatiallyBinnedSpectra
    def write(self, match_DRP=False, overwrite=False):
        """
        Write the hdu object to the file.
        """
        # Convert the spectral arrays in the HDU to a 3D cube and write
        # it
        if match_DRP:
            hdu = self.construct_3d_hdu()
            DAPFitsUtil.write(hdu, str(self.file_path()), overwrite=overwrite, checksum=True,
                              loggers=self.loggers, quiet=self.quiet)
            return
        # Just write the unique (2D) data
        DAPFitsUtil.write(self.hdu, str(self.file_path()), overwrite=overwrite, checksum=True,
                          loggers=self.loggers, quiet=self.quiet) 


    def read(self, ifile=None, strict=True, checksum=False, debug=False):
        """
        Read an existing file with an existing set of emission-line
        models.
        """
        if ifile is None:
            ifile = self.file_path()
        if not ifile.exists():
            raise FileNotFoundError('File does not exist!: {0}'.format(ifile))

        if self.hdu is not None:
            self.hdu.close()

        self.hdu = DAPFitsUtil.read(str(ifile), checksum=checksum)

        # Force load the PAR and FIT extensions because it may be
        # checked by self.construct_3d_hdu() or construct_maps_file in
        # dapfits.py
        try:
            if self.hdu['PAR'].data is None:
                pass
        except:
            warnings.warn('Could not load emission-line parameter data.')
        try:
            if self.hdu['FIT'].data is None:
                pass
        except:
            warnings.warn('Could not load emission-line fit data.')

        # Make sure that the number of emission-lines is set
        _neml = self.hdu['EMLDATA'].data['FLUX'].shape[1]
        if self.neml is not None and _neml != self.neml:
            raise ValueError('Num. of emission lines read does not match emission-line database.')
        if self.neml is None:
            self.neml = _neml

        # Confirm that the internal method is the same as the method
        # that was used in writing the file
        if self.hdu['PRIMARY'].header['ELFKEY'] != self.method['key']:
            if strict:
                raise ValueError('Keywords in header do not match specified method keywords!')
            else:
                warnings.warn('Keywords in header do not match specified method keywords!')
        # TODO: "strict" should also check other aspects of the file to
        # make sure that the details of the method are also the same,
        # not just the keyword

        # Attempt to read the modeling parameters
        if self.method['fitpar'] is not None:
            try:
                self.method['fitpar'].fromheader(self.hdu['PRIMARY'].header)
            except:
                if not self.quiet:
                    warnings.warn('Fit parameter object does not have a fromheader ' \
                                  'attribute.  No parameters head from header.')

        self.nmodels = self.hdu['FLUX'].shape[0]
        unique_bins = numpy.unique(self.hdu['BINID'].data.ravel()) \
                            if self.hdu['PRIMARY'].header['ELREBIN'] else None
        self.missing_models = self._get_missing_models(unique_bins=unique_bins, debug=debug)

    def copy_to_array(self, ext='FLUX', waverange=None, include_missing=False):
        r"""

        Wrapper for :func:`mangadap.util.fitsutil.DAPFitsUtil.copy_to_array`
        specific for :class:`EmissionLineModel`.

        Return a copy of the selected data array.  The array size is
        always :math:`N_{\rm models} \times N_{\rm wavelength}`; i.e., the
        data is always flattened to two dimensions and the unique
        spectra are selected.

        Args:
            ext (:obj:`str`, optional):
                Name of the extension from which to draw the data.  Must be
                allowed for the current :attr:`mode`; see :attr:`data_arrays`.
                Default is ``'FLUX'``.
            waverange (array-like, optional):
                Two-element array with the first and last wavelength to include
                in the computation.  Default is to use the full wavelength
                range.
            include_missing (:obj:`bool`, optional):
                Create an array with a size that accommodates the missing
                models.

        Returns:
            `numpy.ndarray`_: A 2D array with a copy of the data from
            the selected extension.
        """
        return DAPFitsUtil.copy_to_array(self.hdu, ext=ext, allowed_ext=self.spectral_arrays,
                waverange=waverange, missing_bins=self.missing_models if include_missing else None,
                nbins=self.nmodels, unique_bins=DAPFitsUtil.unique_bins(self.hdu['BINID'].data))

    def copy_to_masked_array(self, ext='FLUX', flag=None, waverange=None, include_missing=False):
        r"""
        Wrapper for
        :func:`mangadap.util.fitsutil.DAPFitsUtil.copy_to_masked_array` specific
        for :class:`EmissionLineModel`.

        Return a copy of the selected data array as a masked array.  This is
        functionally identical to :func:`copy_to_array`, except the output
        format is a `numpy.ma.MaskedArray`_.  The pixels that are considered to
        be masked can be specified using `flag`.
        
        Args:
            ext (:obj:`str`, optional):
                Name of the extension from which to draw the data.  Must be
                allowed for the current :attr:`mode`; see :attr:`data_arrays`.
                Default is `'FLUX'`.
            flag (:obj:`str`, :obj:`list`, optional):
                (List of) Flag names that are considered when deciding if a
                pixel should be masked.  The names *must* be a valid bit name as
                defined by :attr:`bitmask`.  If not provided, *ANY* non-zero
                mask bit is omitted.
            waverange (array-like, optional):
                Two-element array with the first and last wavelength to include
                in the computation.  Default is to use the full wavelength
                range.
            include_missing (:obj:`bool`, optional):
                Create an array with a size that accommodates the missing
                models.

        Returns:
            `numpy.ndarray`_: A 2D array with a copy of the data from
            the selected extension.
        """
        return DAPFitsUtil.copy_to_masked_array(self.hdu, ext=ext, mask_ext='MASK', flag=flag,
                bitmask=self.bitmask, allowed_ext=self.spectral_arrays, waverange=waverange,
                missing_bins=self.missing_models if include_missing else None,
                nbins=self.nmodels, unique_bins=DAPFitsUtil.unique_bins(self.hdu['BINID'].data))

    def fill_to_match(self, binid, include_base=False, missing=None):
        """
        Get the emission-line model that matches the input bin ID matrix.  The
        output is a 2D matrix ordered by the bin ID; any skipped index numbers
        in the maximum of the union of the unique numbers in the `binid` and
        `missing` input are masked.

        Use `include_base` to include the baseline in the output emission-line
        models.
        """
        # The input bin id array must have the same shape as self
        if binid.shape != self.spatial_shape:
            raise ValueError('Input bin ID matrix has incorrect shape.')

        # Construct the best-fitting models
        best_fit_model = self.copy_to_array()
        if include_base:
            best_fit_model += self.copy_to_array(ext='BASE')

        # Get the number of output models
        nbins = numpy.amax(binid).astype(int)+1 if missing is None or len(missing) == 0 else \
                    numpy.amax( numpy.append(binid.ravel(), missing) ).astype(int)+1

        # Map the BINID to the spectrum index, assuming bins are sorted
        u, indx, reconstruct = numpy.unique(self['BINID'].data.ravel(), return_index=True,
                                            return_inverse=True)
        u_bin_indx = numpy.arange(u.size)
        if u[0] == -1:
            u_bin_indx -= 1
        _bin_indx = u_bin_indx[reconstruct].reshape(self.spatial_shape)

        # Fill in bins with no models with masked zeros
        emission_lines = numpy.ma.zeros((nbins,best_fit_model.shape[1]), dtype=float)
        emission_lines[:,:] = numpy.ma.masked
        for i,j in zip(binid.ravel(), _bin_indx.ravel()):
            if i < 0 or j < 0:
                continue
            emission_lines[i,:] = best_fit_model[j,:]

        return emission_lines

    def matched_kinematics(self, binid, line_name, redshift=None, dispersion=100.0, constant=False,
                           cz=False, corrected=False, min_dispersion=None, nearest=False,
                           missing=None):
        """
        Return the redshift and velocity dispersion for all the
        spectra based on the emission-line model fit to a single line.

        For spectra the were not fit, use the median (unmasked)
        kinematic measurement if no default value is provided as an
        argument.

        Args:
            binid (`numpy.ndarray`_):
                2D array with the bin ID associate with each spaxel
                in the datacube.
            line (str): Line to use for the kinematics.  Must match one
                of the lines created by :func:`channel_names`.  Function
                will raise an exception if the line_name is None or if
                the channel names cannot be constructed.
            redshift (float): (**Optional**) The default redshift to use
                for spectra without an emission-line fit.  Default is to
                use the median of the unmasked measurements.
            dispersion (float): (**Optional**) The default velocity
                dispersion to use for spectra without an emission-line
                fit.  Default is 100 km/s.
            constant (bool): (**Optional**) Force the function to return
                a constant redshift and dispersion for each spectrum,
                regardless of any fitted kinematics.
            cz (bool): (**Optional**) Return the redshift as cz in km/s,
                as opposed to unitless z.
            corrected (bool): (**Optional**) By default, the returned
                velocity dispersions are the measured values, which will
                include the instrumental resolution.  Setting corrected
                to True will return the corrected dispersions.
            min_dispersion (float): (**Optional**) Impose a minimum
                dispersion.
            nearest (bool): (**Optional**) Instead of the median of the
                results for the spectra that were not fit, use the value
                from the nearest bin.
            missing (list): (**Optional**) Any bin ID numbers missing
                from the input `binid` image needed for constructing the
                output matched data.

        Returns:
            `numpy.ndarray`_: Returns arrays with a redshift (or cz)
            and dispersion for each spectrum.

        Raises:
            TypeError: Raised if the input redshift or dispersion values
                are not single numbers.
        """
        # Check input
        if redshift is not None and not isinstance(redshift, (float,int)):
            raise TypeError('Redshift must be a single number or None.')
        if dispersion is not None and not isinstance(dispersion, (float,int)):
            raise TypeError('Dispersion must be a single number or None.')
        if binid.shape != self.spatial_shape:
            raise ValueError('Input bin ID map must match the spatial shape of the DRP cube.')

        # Get the channel names
        names = self.channel_names()
        if line_name not in names:
            raise ValueError('Line {0} not present in channel list.'.format(line_name))
        vel_channel = numpy.arange(len(names))[names == line_name][0]

        # Get the number of output kinematics
        nbins = numpy.amax(binid).astype(int)+1 if missing is None or len(missing) == 0 else \
                    numpy.amax( numpy.append(binid.ravel(), missing) ).astype(int)+1

        # Mask bad values
        mask = self.bitmask.flagged(self.hdu['EMLDATA'].data['MASK'][:,vel_channel].copy(),
                                    [ 'NO_FIT', 'FIT_FAILED', 'INSUFFICIENT_DATA', 'NEAR_BOUND' ])

        if numpy.all(mask):
            # Unlike StellarContinuumModel.matched_kinematics, the input
            # guesses here always are based on the number of binned
            # spectra.  So if the class is meant to deconstruct these
            # bins, we can't just use the guess kinematics when all the
            # output is masked.  So we'll just use the defaults.
            warnings.warn('All emission line models have been masked!  Using defaults.')
            eml_z = numpy.ma.MaskedArray(numpy.full(self.nmodels,
                                                    (0.0 if redshift is None else redshift),
                                                    dtype=float))
            eml_d = numpy.ma.MaskedArray(numpy.full(self.nmodels,
                                    (100.0 if dispersion is None else dispersion), dtype=float))
        else:
            # Pull out the data
            eml_z = numpy.ma.MaskedArray(self.hdu['EMLDATA'].data['KIN'][:,vel_channel,0].copy(),
                                            mask=mask) / astropy.constants.c.to('km/s').value
            eml_d = numpy.ma.MaskedArray(self.hdu['EMLDATA'].data['KIN'][:,vel_channel,1].copy(),
                                            mask=mask)

            # Apply the sigma corrections if requested.  Values with the
            # correction larger than the measured value will be masked!
            if corrected:
                sigma_corr = numpy.ma.MaskedArray(
                                self.hdu['EMLDATA'].data['SIGMACORR'][:,vel_channel], mask=mask)
                eml_d = numpy.ma.sqrt(numpy.square(eml_d) - numpy.square(sigma_corr))

        # Set the default values to use when necessary
        _redshift = numpy.ma.median(eml_z) if redshift is None else redshift
        _dispersion = numpy.ma.median(eml_d) if dispersion is None else dispersion
        if _dispersion is numpy.ma.masked:
            _dispersion = 100.0
        if min_dispersion is not None and _dispersion < min_dispersion:
            _dispersion = min_dispersion

        # Just return the single value
        if constant:
            eml_z = numpy.full(nbins, _redshift * astropy.constants.c.to('km/s').value
                                        if cz else _redshift, dtype=float)
            eml_d = numpy.full(nbins, _dispersion, dtype=float)
            return eml_z, eml_d

        if nearest:
            replace = numpy.ma.getmaskarray(eml_z) | numpy.ma.getmaskarray(eml_d)
            if numpy.all(replace):
                # All are bad so replace with _redshift and _dispersion
                eml_z = numpy.full(eml_z.shape, _redshift, dtype=float)
                eml_d = numpy.full(eml_d.shape, _dispersion, dtype=float)
            elif numpy.any(replace):
                # Fill masked values with the nearest datum
                best_fit_kinematics = numpy.ma.append([eml_z], [eml_d], axis=0).T
                if self.method['fitpar']['deconstruct_bins'] != 'ignore':
                    indx = self['BINID'].data.ravel() > -1
                    coo = self.binned_spectra.rdxqa['SPECTRUM'].data['SKY_COO'][indx,:]
                else:
                    valid_bins = numpy.unique(self['BINID'].data)
                    if valid_bins[0] == -1:
                        valid_bins = valid_bins[1:]
                    coo = self.binned_spectra['BINS'].data['SKY_COO'][valid_bins,:]
                replace = eml_z.mask | eml_d.mask
                kinematics = replace_with_data_from_nearest_coo(coo, best_fit_kinematics, replace)
                eml_z = kinematics[:,0]
                eml_d = kinematics[:,1]
        else:
            # Fill any masked values with the single estimate
            eml_z = eml_z.filled(_redshift)
            eml_d = eml_d.filled(_dispersion)

        # Map the BINID to the spectrum index, assuming bins are sorted
        # and that the BINID map has -1 BINID values
        u, indx, reconstruct = numpy.unique(self['BINID'].data.ravel(), return_index=True,
                                            return_inverse=True)
#        u_bin_indx = numpy.arange(len(u))-1
        u_bin_indx = numpy.arange(u.size)
        if u[0] == -1:
            u_bin_indx -= 1
        _bin_indx = u_bin_indx[reconstruct].reshape(self.spatial_shape)

        # Match the kinematics to the output bin ID map
        _eml_z = numpy.ma.masked_all(nbins, dtype=float)
        _eml_d = numpy.ma.masked_all(nbins, dtype=float)
        for i,j in zip(binid.ravel(), _bin_indx.ravel()):
            if i < 0 or j < 0:
                continue
            _eml_z[i] = eml_z[j]
            _eml_d[i] = eml_d[j]

        eml_z = _eml_z.filled(_redshift)
        eml_d = _eml_d.filled(_dispersion)
        # Convert to cz velocities (km/s)
        if cz:
            eml_z *= astropy.constants.c.to('km/s').value
        return eml_z, eml_d

    def fill_continuum_to_match(self, binid, replacement_templates=None, redshift_only=False,
                                deredshift=False, corrected_dispersion=False, missing=None):
        """
        Get the emission-line continuum model, if possible, that matches
        the input bin ID matrix.  The output is a 2D matrix ordered by
        the bin ID; any skipped index numbers in the maximum of the
        union of the unique numbers in the `binid` and `missing` input
        are masked.

        Use `replacement_templates` only if the number of templates is
        identical to the number used during the fitting procedure.  Use
        `redshift_only` to produce the best-fitting model without and
        velocity dispersion.  Use `corrected_dispersion` to produce the
        model using the corrected velocity dispersion.

        """
        if self.stellar_continuum is None:
            raise ValueError('EmissionLineModel object has no associated StellarContinuumModel.')
        if self.method['fitclass'] is None:
            raise ValueError('No class object available for constructing the model!')
        if not callable(self.method['fitclass'].construct_continuum_models):
            raise AttributeError('Provided fit class object has no callable ' \
                                 '\'construct_continuum_models\' attribute!')
        if corrected_dispersion and 'NEAREST_BIN' not in self.hdu['FIT'].columns.names:
            raise KeyError('Nearest stellar continuum bin not set in fit parameters.  Cannot '
                           'apply dispersion corrections.')

        # The input bin id array must have the same shape as self
        if binid.shape != self.spatial_shape:
            raise ValueError('Input bin ID matrix has incorrect shape.')

        # Pull out the dispersion corrections, if requested
        dispersion_corrections = None
        if corrected_dispersion:
            nsc = self.stellar_continuum['PAR'].data['BINID'].size
            indx = [ numpy.arange(nsc)[self.stellar_continuum['PAR'].data['BINID'] == nb][0]
                        for nb in self.hdu['FIT'].data['NEAREST_BIN'] ]
            dispersion_corrections = self.stellar_continuum['PAR'].data['SIGMACORR_SRES'][indx]

        # Get the continuum models
        best_fit_continuum = self.construct_continuum_models(
                                            replacement_templates=replacement_templates,
                                            deredshift=deredshift, redshift_only=redshift_only,
                                            dispersion_corrections=dispersion_corrections)

        # Get the number of output continuua
        nbins = numpy.amax(binid).astype(int)+1 if missing is None or len(missing) == 0 else \
                    numpy.amax( numpy.append(binid.ravel(), missing) ).astype(int)+1

        # Map the BINID to the spectrum index, assuming bins are sorted
        # and that the BINID map has -1 BINID values
        u, indx, reconstruct = numpy.unique(self['BINID'].data.ravel(), return_index=True,
                                            return_inverse=True)
#        u_bin_indx = numpy.arange(len(u))-1
        u_bin_indx = numpy.arange(u.size)
        if u[0] == -1:
            u_bin_indx -= 1
        _bin_indx = u_bin_indx[reconstruct].reshape(self.spatial_shape)

        # Fill in bins with no models with masked zeros
        continuum = numpy.ma.zeros((nbins,best_fit_continuum.shape[1]), dtype=float)
        continuum[:,:] = numpy.ma.masked
        for i,j in zip(binid.ravel(), _bin_indx.ravel()):
            if i < 0 or j < 0:
                continue
            continuum[i,:] = best_fit_continuum[j,:]

        return continuum

    def construct_continuum_models(self, replacement_templates=None, redshift_only=False,
                                   deredshift=False, dispersion_corrections=None):

        if self.stellar_continuum is None:
            raise ValueError('EmissionLineModel object has no associated StellarContinuumModel.')

        if self.method['fitclass'] is None:
            raise ValueError('No class object available for constructing the model!')
        if not callable(self.method['fitclass'].construct_continuum_models):
            raise AttributeError('Provided fit class object has no callable ' \
                                 '\'construct_continuum_models\' attribute!')
        if replacement_templates is not None \
                and not isinstance(replacement_templates, TemplateLibrary):
            raise TypeError('Provided template replacements must have type TemplateLibrary.')

        # Only the selected models are constructed, others are masked
        select = numpy.invert(self.bitmask.flagged(self.hdu['FIT'].data['MASK'],
                                            flag=['NO_FIT', 'FIT_FAILED', 'INSUFFICIENT_DATA']))

        # Get the template library
        templates = replacement_templates
        if templates is None or not isinstance(templates, TemplateLibrary):
            try:
                templates = self.method['fitpar'].templates
#
#                bins_to_fit = EmissionLineFit.select_binned_spectra_to_fit(self.binned_spectra,
#                                            minimum_snr=self.method['fitpar']['minimum_snr'],
#                                            stellar_continuum=self.stellar_continuum)
#                z = numpy.mean(self.method['fitpar']['guess_redshift'][bins_to_fit])
#                templates = self.method['fitclass'].get_stellar_templates(self.method['fitpar'],
#                                                                          self.binned_spectra.cube,
#                                                                          z=z,
#                                                                          loggers=self.loggers,
#                                                                          quiet=self.quiet)[0]
            except:
                templates = self.stellar_continuum.method['fitpar'].templates

        return self.method['fitclass'].construct_continuum_models(
                                        self.method['fitpar'].emldb, templates['WAVE'].data,
                                        templates['FLUX'].data, self.hdu['WAVE'].data,
                                        self.hdu['FLUX'].data.shape, self.hdu['FIT'].data,
                                        select=select, redshift_only=redshift_only,
                                        deredshift=deredshift,
                                        dispersion_corrections=dispersion_corrections, dvtol=1e-9)

    # TODO: Use the EmissionMomentsDB.channel_names function!
    def channel_names(self):
        if self.method['fitpar'].emldb is None:
            raise ValueError('Channel names undefined because line database is not available.')
        
        return numpy.array([ '{0}-{1}'.format(n,int(w)) \
                        for n,w in zip(self['PAR'].data['NAME'],
                                       self['PAR'].data['RESTWAVE']) ])

    def fit_figures_of_merit(self):
        """
        Use the fitting class to set the figures-of-merit of the full
        fit.

        This is primarily used to set the output data for the MAPS file.
        
        Returns:
            :obj:`tuple`: Returns (1) a list of names to assign each
            column (length is NFOM), (2) the units of the data in
            this column (length is NFOM), and (3) a 2D array with the
            figures of merit for each spectrum (shape is NSPEC x
            NFOM). If the fitting class used does not have a
            `fit_figures_of_merit` function, the name is set to
            'null', the unit is an empty string, and the data is a
            (NMOD,1) array of zeros.
        """
        try:
            names, units, fom = self.method['fitclass'].fit_figures_of_merit(self.hdu['FIT'].data)
        except:
            names = ['null']
            units = ['']
            fom = numpy.zeros((self.nmodels,1), dtype=float)
        return names, units, fom
        





# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
Provides a set of utility function that should **only be used for the
survey-level execution of the DAP**.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import subprocess
import os


def product_version(product='mangadap', simple=False):
    """
    Get currently loaded SDSS product version by parsing the output of
    the system call to {$product}_version.

    Args:
        product (str): (Optional) The name of the product
        simple (bool): (Optional) If True, only the first element of the
            reported version is returned.

    Returns:
        str : Version identifier
    """
    try:
        version = subprocess.check_output('%s_version' % product, shell=True)
        if type(version) is bytes:
            version = version.decode('utf-8')
        version = version.split(' ')[0].rstrip('\n') if simple else version.rstrip('\n')
    except Exception as e:
#        print_frame('Exception')
#        print(e)
        version = None

    return version


def module_version(product='mangadap'):
    """
    Return the version of the specified product among the currently
    loaded modules.

    Args:
        product (str): (Optional) The name of the product

    Returns:
        str : Version identifier
    """
    try:
        modules = os.environ['LOADEDMODULES']
    except:
#        print_frame('Exception')
#        modules = None
        return None
    # TODO: Re-raise the exception?
  
    # Parse the loaded version(s) of product
    versions = [module.split('/')[1] for module in modules.split(':')
                    if module.split('/')[0]==product]

    # If there is more than one version or no versions return None
    if len(versions) != 1:
        if len(versions) > 1:
            print('Multiple versions found for module {0}'.format(product))
        else:
            print('Module {0} is not loaded'.format(product))
        return None

    # Return the version
    return versions[0]








#!/usr/bin/env python3

import time
import warnings
import numpy
import os
import astropy.constants
import pdb

from mangadap.drpfits import DRPFits
from mangadap.proc.reductionassessments import ReductionAssessment
from mangadap.proc.spectralstack import SpectralStackPar, SpectralStack
from mangadap.proc.spatiallybinnedspectra import SpatiallyBinnedSpectra, SpatiallyBinnedSpectraDef
from mangadap.proc.stellarcontinuummodel import StellarContinuumModel
from mangadap.proc.emissionlinemoments import EmissionLineMoments
from mangadap.proc.emissionlinemodel import EmissionLineModel
from mangadap.proc.spectralindices import SpectralIndices
from mangadap.dapfits import construct_maps_file, construct_cube_file

#-----------------------------------------------------------------------------
#Based on KBW's test_new_binning_scheme.py

class SquareBinning():

    """
    Perform binning of full cube in square apertures
    Length of aperture side is given in arcsec with binsz

    Args:
        binsz (array-like): desired bin size

    """
    def __init__(self, binsz):
        self.binsz = binsz


    def bin_spaxels(self, x, y, par=None):
        _x = numpy.asarray(x)
        if len(_x.shape) != 1:
            raise ValueError('On-sky coordinates must be one-dimensional.')
        nspaxels = _x.size
        if len(y) != nspaxels:
            raise ValueError('Input coordinates are of different lengths.')
        _y = numpy.asarray(y)

        binid = numpy.full(nspaxels, -1, dtype=int)
        binsz = self.binsz
        minx = numpy.min(_x)-0.25
        maxx = numpy.max(_x)+0.25
        miny = numpy.min(_y)-0.25
        maxy = numpy.max(_y)+0.25

        ctbin = 0

        # Start from the center, work out to edge of each quadrant

        # Quadrant 1
        nslicex = numpy.ceil((maxx) / binsz)
        nslicey = numpy.ceil((maxy) / binsz)

        x_lo = numpy.linspace(0.0,(nslicex*binsz),nslicex+1.0)
        y_lo = numpy.linspace(0.0,(nslicey*binsz),nslicey+1.0)

        # Find which spaxels land in each aperture
        for xi in x_lo:
            for yj in y_lo:
                indx = (_x >= xi) & (_x < xi+binsz) & (_y >= yj) & (_y < yj+binsz)
                if any(indx):
                    binid[indx] = ctbin
                    ctbin = ctbin + 1


        # Quadrant 2
        nslicex = numpy.ceil((maxx) / binsz)
        nslicey = numpy.floor((miny) / binsz)

        x_lo = numpy.linspace(0.0, (nslicex * binsz), nslicex + 1.0)
        y_lo = numpy.linspace(0.0, (nslicey * binsz), numpy.abs(nslicey) + 1.0)

        # Find which spaxels land in each aperture
        for xi in x_lo:
            for yj in y_lo:
                indx = (_x >= xi) & (_x < xi + binsz) & (_y <= yj) & (_y > yj - binsz)
                if any(indx):
                    binid[indx] = ctbin
                    ctbin = ctbin + 1

        # Quadrant 3
        nslicex = numpy.floor((minx) / binsz)
        nslicey = numpy.ceil((maxy) / binsz)

        x_lo = numpy.linspace(0.0, (nslicex * binsz), numpy.abs(nslicex) + 1.0)
        y_lo = numpy.linspace(0.0, (nslicey * binsz), numpy.abs(nslicey) + 1.0)

        # Find which spaxels land in each aperture
        for xi in x_lo:
            for yj in y_lo:
                indx = (_x <= xi) & (_x > xi - binsz) & (_y >= yj) & (_y < yj + binsz)
                if any(indx):
                    binid[indx] = ctbin
                    ctbin = ctbin + 1

        # Quadrant 4
        nslicex = numpy.floor((minx) / binsz)
        nslicey = numpy.floor((miny) / binsz)

        x_lo = numpy.linspace(0.0, (nslicex * binsz), numpy.abs(nslicex) + 1.0)
        y_lo = numpy.linspace(0.0, (nslicey * binsz), numpy.abs(nslicey) + 1.0)

        # Find which spaxels land in each aperture
        for xi in x_lo:
            for yj in y_lo:
                indx = (_x <= xi) & (_x > xi - binsz) & (_y <= yj) & (_y > yj - binsz)
                if any(indx):
                    binid[indx] = ctbin
                    ctbin = ctbin + 1


        #pdb.set_trace()
        return binid




#-----------------------------------------------------------------------------
if __name__ == '__main__':
    t = time.clock()

    # Set the plate, ifu, and initial velocity/redshift
    binsz = 2.5        # Desired bin size in arcsec
    plate = 7443
    ifu = 1902
    #vel = 8675.5
    #nsa_redshift = vel/astropy.constants.c.to('km/s').value
    #nsa_redshift = 0.0280161   # for 7443-6102
    #nsa_redshift = 0.020478
    nsa_redshift = 0.0189259
    vel = nsa_redshift * astropy.constants.c.to('km/s').value

    analysis_path = os.getcwd()

    # Read the DRP LOGCUBE file
    drpf = DRPFits(plate, ifu, 'CUBE', read=True)


    # Set up binning function
    sqbin = SquareBinning(binsz)


    # Calculate the S/N and coordinates
    rdxqa = ReductionAssessment('SNRG', drpf, analysis_path=analysis_path)
    x = rdxqa.hdu['SPECTRUM'].data['SKY_COO'][:, 0]
    y = rdxqa.hdu['SPECTRUM'].data['SKY_COO'][:, 1]
    fgdpix = rdxqa.hdu['SPECTRUM'].data['FGOODPIX'] > 0.8

    #binid = sqbin.bin_spaxels(x[fgdpix], y[fgdpix], par=None)
    #pdb.set_trace()

    # Setup the stacking operations
    stackpar = SpectralStackPar('mean',  # Operation for stack
                                False,  # Apply a velocity registration
                                None,  # Velocity offsets for registration
                                'channels',  # Covariance mode and parameters
                                SpectralStack.parse_covariance_parameters('channels', 11),
                                True,  # Propagate the LSF through the stacking
                                True)  # Use pre-pixelized LSF (KHRR added this)
    stacker = SpectralStack()

    # Create a new binning method
    binning_method = SpatiallyBinnedSpectraDef('5x5n',  # Key for binning method
                                               'ODonnell',  # Galactic reddening function to use
                                               3.1,  # Rv for Galactic reddening
                                               0.0,  # Minimum S/N to include
                                               None,  # Object with binning parameters
                                               None,  # Binning class instance
                                               sqbin.bin_spaxels,  # Binning function
                                               stackpar,  # Object with stacking parameters
                                               stacker,  # Stacking class instance
                                               stacker.stack_DRPFits,  # Stacking function
                                               'spaxel',  # Type of LSF characterization to use
                                               True)  # Use pre-pixelized LSF (KHRR added this)

    # Bin the spectra using the new binning method
    binned_spectra = SpatiallyBinnedSpectra('5x5n',  # Key for binning method
                                            drpf,  # DRP data to bin
                                            rdxqa,  # Cube coordinates and S/N assessments
                                            method_list=binning_method,  # Methods to select from
                                            analysis_path=analysis_path)

    # The rest of this is just a single execution of the remaining
    # analysis steps in
    # $MANGADAP_DIR/python/mangadap/survey/manga_dap.py , with some
    # simplifications

    stellar_continuum = StellarContinuumModel('GAU-MILESHC', binned_spectra, guess_vel=vel,
                                              guess_sig=100., analysis_path=analysis_path)

    emission_line_moments = EmissionLineMoments('EMOMM', binned_spectra,
                                                stellar_continuum=stellar_continuum,
                                                redshift=nsa_redshift, analysis_path=analysis_path)
    #emission_line_moments = None

    emission_line_model = EmissionLineModel('EFITM', binned_spectra,
                                            stellar_continuum=stellar_continuum,
                                            redshift=nsa_redshift, dispersion=100.0,
                                            analysis_path=analysis_path)
    #emission_line_model = None

    #    spectral_indices = SpectralIndices('INDXEN', binned_spectra, redshift=nsa_redshift,
    #                                       stellar_continuum=stellar_continuum,
    #                                       emission_line_model=emission_line_model,
    #                                       analysis_path=analysis_path)
    spectral_indices = None

    construct_maps_file(drpf, rdxqa=rdxqa, binned_spectra=binned_spectra,
                        stellar_continuum=stellar_continuum,
                        emission_line_moments=emission_line_moments,
                        emission_line_model=emission_line_model,
                        spectral_indices=spectral_indices, nsa_redshift=nsa_redshift,
                        analysis_path=analysis_path)

    construct_cube_file(drpf, binned_spectra=binned_spectra, stellar_continuum=stellar_continuum,
                        emission_line_model=emission_line_model, analysis_path=analysis_path)

    print('Elapsed time: {0} seconds'.format(time.clock() - t))



    #pdb.set_trace()